import cProfile
import pstats
from functools import wraps
from io import StringIO
from logging import debug
from typing import Callable


def profile(func: Callable):
    """
    Decorator to profile a function.

    This is debug tool, so its usage should not be committed. Decorate wit it, profile code, remove usage.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        retval = func(*args, **kwargs)
        pr.disable()
        s = StringIO()
        sortby = "cumulative"
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        debug(f"Profile stats for {func.__name__}")
        debug(s.getvalue())
        return retval

    return wrapper



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_LIFECYCLE_MAN",
    description="GRO_LIFECYCLE_MAN_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_GENETIC_ENG"],
    unlock=[
        Item(type=UnlockShipPart, name="CO_SUSPEND_ANIM_POD"),
        Item(type=UnlockPolicy, name="PLC_NO_GROWTH"),
    ],
    graphic="icons/tech/lifecycle_manipulation.png",
)



from focs._effects import EffectsGroup, Focus, OwnedBy, Planet, SetMaxStockpile, Source, Value
from focs._tech import *

Tech(
    name="PRO_PREDICTIVE_STOCKPILING",
    description="PRO_PREDICTIVE_STOCKPILING_DESC",
    short_description="IMPERIAL_STOCKPILE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    effectsgroups=[
        # Set initial meters
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Focus(type=["FOCUS_STOCKPILE"]),
            effects=SetMaxStockpile(value=Value + 1),
        )
    ],
)



from stub_generator.stub_generator.rtype._function_mapping import update_function_rtype
from stub_generator.stub_generator.rtype._method_mapping import update_method_rtype
from stub_generator.stub_generator.rtype._property_mapping import update_property_rtype



from focs._effects import (
    EffectsGroup,
    Full,
    HasSpecies,
    IsSource,
    OwnedBy,
    Planet,
    SetDetection,
    SetVisibility,
    Ship,
    Source,
    Unowned,
    Value,
)
from macros.priorities import DEFAULT_PRIORITY

BAD_DETECTION = [
    EffectsGroup(
        description="BAD_DETECTION_DESC",
        scope=IsSource,
        activation=~Ship,
        effects=SetDetection(value=Value - 20),
    ),
    EffectsGroup(scope=IsSource, activation=Ship, effects=SetDetection(value=Value - 9)),
]


GOOD_DETECTION = [
    EffectsGroup(
        description="GOOD_DETECTION_DESC",
        scope=IsSource,
        effects=SetDetection(value=Value + 25),
    ),
]

GREAT_DETECTION = [
    EffectsGroup(
        description="GREAT_DETECTION_DESC",
        scope=IsSource,
        effects=SetDetection(value=Value + 50),
    ),
]


ULTIMATE_DETECTION = [
    EffectsGroup(
        description="ULTIMATE_DETECTION_DESC",
        scope=IsSource,
        effects=SetDetection(value=Value + 100),
    )
]


def NATIVE_PLANETARY_DETECTION(detection):
    return EffectsGroup(
        scope=IsSource,
        activation=Planet() & Unowned,
        accountinglabel="NATIVE_PLANETARY_DETECTION_LABEL",
        priority=DEFAULT_PRIORITY,
        effects=SetDetection(value=Value + detection),
    )


def COMMUNAL_VISION(species: str):
    return EffectsGroup(
        description="COMMUNAL_VISION_DESC",
        scope=Planet()
        & HasSpecies(name=[species])
        & ~OwnedBy(empire=Source.Owner),  # would be redundant to re-assign visbility to own planets
        activation=Planet() & ~Unowned,
        effects=SetVisibility(empire=Source.Owner, visibility=Full),
    )



from focs._effects import (
    EffectsGroup,
    GenerateSitRepMessage,
    IsSource,
    Population,
    SetPopulation,
    Source,
    Target,
    Value,
)
from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import GASEOUS_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import VERY_BAD_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS, NO_OFFENSE_TROOPS

Species(
    name="SP_THENIAN",
    description="SP_THENIAN_DESC",
    gameplay_description="SP_THENIAN_GAMEPLAY_DESC",
    native=True,
    tags=[
        "GASEOUS",
        "NO_TERRAFORM",
        "ARTISTIC",
        "AVERAGE_POPULATION",
        "AVERAGE_HAPPINESS",
        "VERY_BAD_SUPPLY",
        "NO_INDUSTRY",
        "GOOD_STEALTH",
        "NO_OFFENSE_TROOPS",
        "PEDIA_GASEOUS_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_GROWTH",
    likes=[
        "FOCUS_RESEARCH",  # arty farty
    ],
    dislikes=[
        "BLD_GAS_GIANT_GEN",  # messes with their habitat
        "BLD_SPACE_ELEVATOR",  # see above
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_THENIAN"),
        *VERY_BAD_SUPPLY,
        *NO_OFFENSE_TROOPS,
        *GOOD_DEFENSE_TROOPS,
        *GOOD_STEALTH,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        NARROW_EP,
        EffectsGroup(
            scope=IsSource & Population(low=Source.TargetPopulation - 0.01),  # when maximum population is reached
            effects=[
                SetPopulation(value=Value * 0.5),  # then dieoff to half
                GenerateSitRepMessage(  # and tell the owner it happened
                    message="EFFECT_THENIAN_NATURAL_DIEOFF",
                    label="EFFECT_THENIAN_NATURAL_DIEOFF_LABEL",
                    icon="icons/sitrep/colony_starvation.png",
                    parameters={"tag": "planet", "data": Target.ID},
                    empire=Target.Owner,
                ),
            ],
        ),
    ],
    environments=GASEOUS_NARROW_EP,
    graphic="icons/species/thenian.png",
)



from focs._effects import (
    BlackHole,
    Blue,
    EffectsGroup,
    IsSource,
    NamedReal,
    Neutron,
    NoStar,
    OwnedBy,
    OwnerHasTech,
    Planet,
    SetDetection,
    SetEmpireMeter,
    Ship,
    Source,
    Star,
    Value,
    White,
    Yellow,
)
from focs._tech import *

Tech(
    name="SPY_DETECT_1",
    description="SPY_DETECT_1_DESC",
    short_description="DETECTION_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=1,
    researchturns=1,
    researchable=False,
    tags=["PEDIA_SPY_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            activation=~OwnerHasTech(name="SPY_DETECT_2")
            & ~OwnerHasTech(name="SPY_DETECT_3")
            & ~OwnerHasTech(name="SPY_DETECT_4")
            & ~OwnerHasTech(name="SPY_DETECT_5"),
            effects=SetDetection(value=Value + NamedReal(name="SPY_DETECT_1_RANGE", value=50.0)),
        ),
        EffectsGroup(
            scope=IsSource,
            activation=~OwnerHasTech(name="SPY_DETECT_2")
            & ~OwnerHasTech(name="SPY_DETECT_3")
            & ~OwnerHasTech(name="SPY_DETECT_4")
            & ~OwnerHasTech(name="SPY_DETECT_5"),
            effects=SetEmpireMeter(empire=Source.Owner, meter="METER_DETECTION_STRENGTH", value=Value + 10),
        ),
        EffectsGroup(
            scope=(Planet() | Ship) & OwnedBy(empire=Source.Owner) & Star(type=[Blue, Neutron]),
            accountinglabel="SPY_DETECT_STELLAR_INTERFERENCE",
            effects=SetDetection(value=Value - 15),
        ),
        EffectsGroup(
            scope=(Planet() | Ship) & OwnedBy(empire=Source.Owner) & Star(type=[White]),
            accountinglabel="SPY_DETECT_STELLAR_INTERFERENCE",
            effects=SetDetection(value=Value - 10),
        ),
        EffectsGroup(
            scope=(Planet() | Ship) & OwnedBy(empire=Source.Owner) & Star(type=[Yellow]),
            accountinglabel="SPY_DETECT_STELLAR_INTERFERENCE",
            effects=SetDetection(value=Value - 5),
        ),
        EffectsGroup(
            scope=(Planet() | Ship) & OwnedBy(empire=Source.Owner) & Star(type=[BlackHole]),
            accountinglabel="SPY_DETECT_LENSING",
            effects=SetDetection(value=Value + 5),
        ),
        EffectsGroup(
            scope=(Planet() | Ship) & OwnedBy(empire=Source.Owner) & Star(type=[NoStar]),
            accountinglabel="SPY_DETECT_CLEAN_STELLAR_BACKGROUND",
            effects=SetDetection(value=Value + 10),
        ),
    ],
)



from ship_design._additional_specification import AdditionalSpecifications
from ship_design._design_stats import DesignStats
from ship_design._ship_design_cache import ShipDesignCache, build_cache_key



from focs._effects import (
    Armed,
    DesignHasPart,
    EffectsGroup,
    IsSource,
    NamedReal,
    SetMaxDamage,
    SetMaxSecondaryStat,
    Ship,
    Value,
)
from macros.misc import FIGHTER_DAMAGE_FACTOR, SHIP_WEAPON_DAMAGE_FACTOR


def _weapon(*, tag: str, tier_0: int, tier_1: int, tier_2: int, tier_3: int, tier_4: int, hangar: float):
    return [
        EffectsGroup(
            description=f"{tag}_WEAPONS_DESC",
            scope=IsSource,
            activation=Ship
            & Armed
            & (
                DesignHasPart(name="SR_WEAPON_1_1")
                | DesignHasPart(name="SR_WEAPON_2_1")
                | DesignHasPart(name="SR_WEAPON_3_1")
                | DesignHasPart(name="SR_WEAPON_4_1")
            ),
            effects=[
                SetMaxDamage(partname="SR_WEAPON_1_1", value=Value + SHIP_WEAPON_DAMAGE_FACTOR * tier_1),
                SetMaxDamage(partname="SR_WEAPON_2_1", value=Value + SHIP_WEAPON_DAMAGE_FACTOR * tier_2),
                SetMaxDamage(partname="SR_WEAPON_3_1", value=Value + SHIP_WEAPON_DAMAGE_FACTOR * tier_3),
                SetMaxDamage(partname="SR_WEAPON_4_1", value=Value + SHIP_WEAPON_DAMAGE_FACTOR * tier_4),
            ],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=Ship
            & Armed
            & (
                DesignHasPart(name="FT_HANGAR_2")
                | DesignHasPart(name="FT_HANGAR_3")
                | DesignHasPart(name="FT_HANGAR_4")
            ),
            # TODO leave a comment why value multiplier are the same, but weaporns have different
            effects=[
                SetMaxSecondaryStat(
                    partname="FT_HANGAR_2",
                    value=Value
                    + NamedReal(name=tag + "_PILOT_FIGHTER_DAMAGE_BONUS", value=FIGHTER_DAMAGE_FACTOR * hangar),
                ),
                SetMaxSecondaryStat(partname="FT_HANGAR_3", value=Value + FIGHTER_DAMAGE_FACTOR * hangar),
                SetMaxSecondaryStat(partname="FT_HANGAR_4", value=Value + FIGHTER_DAMAGE_FACTOR * hangar),
            ],
        ),
        EffectsGroup(
            scope=IsSource,
            activation=Ship & Armed & DesignHasPart(name="SR_WEAPON_0_1"),
            effects=[
                SetMaxSecondaryStat(partname="SR_WEAPON_0_1", value=Value + tier_0),
            ],
        ),
    ]


BAD_WEAPONS = _weapon(tag="BAD", tier_0=-1, tier_1=-1, tier_2=-2, tier_3=-3, tier_4=-5, hangar=-1.0)
GOOD_WEAPONS = _weapon(tag="GOOD", tier_0=1, tier_1=1, tier_2=2, tier_3=3, tier_4=5, hangar=1.5)
GREAT_WEAPONS = _weapon(tag="GREAT", tier_0=2, tier_1=2, tier_2=4, tier_3=6, tier_4=10, hangar=3.0)
ULTIMATE_WEAPONS = _weapon(tag="ULTIMATE", tier_0=3, tier_1=3, tier_2=6, tier_3=9, tier_4=15, hangar=4.5)



from collections.abc import Sequence

from common.fo_typing import EmpireId, PlayerId, Turn
from stub_generator.stub_generator.collection_classes import make_type
from stub_generator.stub_generator.rtype.mapper import Mapper
from stub_generator.stub_generator.rtype.utils import get_name_for_mapping

_rtypes_map = Mapper(
    "global_functions",
    {
        "currentTurn": Turn,
        "empireID": EmpireId,
        "allEmpireIDs": Sequence[EmpireId],
        "empirePlayerID": PlayerId,
        "getOptionsDBOptionBool": bool,
        "getOptionsDBOptionDouble": float,
        "getOptionsDBOptionInt": int,
        "getOptionsDBOptionStr": str,
    },
)


def update_function_rtype(name: str, rtype: str) -> str:
    key = name
    if key in _rtypes_map:
        return get_name_for_mapping(_rtypes_map[key])
    else:
        return make_type(rtype)



"""
This module extend freeOrionAIInterface from python side.
Check that you have import of this module from FreeOrionAI.py


Adding to string:
=================

Adds object string methods to enable output as
{unique_object_prefix}{id}<{name}>
"P(1)<Sirius>" would be for the planet with id=1 and name=Sirius
This will help to track objects around logs.

str vs repr:
============

If you print item or %s it __str__ will be called if you repr, print list of items or %r __repr__ will be called.

Code organisation:
==================

keep fo.<attr> in same block

def planet_to_string():
    pass


def planet_repr():
    pass

fo.planet.__str__ = planet_to_string
fo.planet.__repr__ = planet_repr


Logger:
=======
wrap any fo function or method to logger to get more info about it.
don't commit logger wrapped function to repo.

"""
import freeOrionAIInterface as fo
import inspect
import os
from functools import wraps
from itertools import zip_longest
from logging import debug

from ._freeorion_tools import dict_from_map, dict_from_map_recursive

PLANET = "P"
SYSTEM = "S"
FLEET = "F"
SHIP_DESIGN = "D"
EMPIRE = "E"


def to_dict(method):
    @wraps(method)
    def wrapper(*args):
        return dict_from_map(method(*args))

    return wrapper


def to_dict_recursive(method):
    @wraps(method)
    def wrapper(*args):
        return dict_from_map_recursive(method(*args))

    return wrapper


def to_str(prefix, id, name):
    return f"{prefix}_{id}<{name}>"


def patch_interface():
    fo.universe.getVisibilityTurnsMap = to_dict(fo.universe.getVisibilityTurnsMap)
    fo.empire.supplyProjections = to_dict(fo.empire.supplyProjections)
    fo.GameRules.getRulesAsStrings = to_dict(fo.GameRules.getRulesAsStrings)
    fo.universe.statRecords = to_dict_recursive(fo.universe.statRecords)

    fo.to_str = to_str

    def system_to_string(system):
        return to_str(SYSTEM, system.systemID, system.name)

    fo.system.__repr__ = system_to_string

    def design_to_string(design):
        return to_str(SHIP_DESIGN, design.id, design.name)

    fo.shipDesign.__repr__ = design_to_string

    def planet_to_string(planet):
        return to_str(PLANET, planet.id, planet.name)

    fo.planet.__repr__ = planet_to_string

    def fleet_to_string(fleet):
        return to_str(FLEET, fleet.id, fleet.name)

    fo.fleet.__repr__ = fleet_to_string

    def empire_to_string(empire):
        return to_str(EMPIRE, empire.empireID, empire.name)

    fo.empire.__repr__ = empire_to_string

    def set_to_string(val):
        return "{%s}" % ", ".join(map(str, val))

    fo.StringSet.__str__ = set_to_string
    fo.IntSet.__str__ = set_to_string

    def int_int_map_to_string(int_int_map):
        return str(dict_from_map(int_int_map))

    fo.IntIntMap.__str__ = int_int_map_to_string


def logger(callable_object, argument_wrappers=None):
    """
    Wrapper to log call of any callable (function or method). It is useful for inspecting API calls.

    Example:
        fo.issueEnqueueShipProductionOrder = logger(fo.issueEnqueueShipProductionOrder,
                                                    argument_wrappers=[
                                                        lambda x: fo.getShipDesign(x).name,
                                                        lambda x: fo.getUniverse().getPlanet(x)
                                                        ]
                                                    )
    Output:
        ProductionAI.py:1352 issueEnqueueShipProductionOrder(UN Cerberus Mk. 5, P_22<Invincible Albans I>) -> 1

    :param callable_object: object to be wrapped
    :param argument_wrappers: list of function to print arguments in different way when they we passed.
     for example ``lambda x: fo.getUniverse().getPlanet(x)`` for planet id.
    :return:
    """

    def inner(*args, **kwargs):
        arguments = [str(wrapper(arg) if wrapper else arg) for arg, wrapper in zip_longest(args, argument_wrappers)]
        arguments.extend("{}={}".format(*item) for item in kwargs.items())
        res = callable_object(*args, **kwargs)
        frame = inspect.currentframe().f_back
        debug(
            "%s:%s %s(%s) -> %s",
            os.path.basename(frame.f_code.co_filename),
            frame.f_lineno,
            callable_object.__name__,
            ", ".join(arguments),
            res,
        )
        return res

    return inner



from focs._effects import (
    ContainedBy,
    Contains,
    EffectsGroup,
    Focus,
    GenerateSitRepMessage,
    HasTag,
    Monster,
    OwnedBy,
    OwnerHasTech,
    Planet,
    Random,
    SetOwner,
    Ship,
    Source,
    System,
    Target,
    VisibleToEmpire,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_PSY_DOM",
    description="LRN_PSY_DOM_DESC",
    short_description="MIND_CONTROL_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=1050 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_UNIF_CONC", "LRN_TIME_MECH"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship
            & ~OwnedBy(empire=Source.Owner)
            & VisibleToEmpire(empire=Source.Owner)
            & Random(probability=0.1)
            & ~Monster
            &
            # @content_tag{TELEPATHIC} For ships with this tag, prevents chance of ownership loss to suitable planet
            ~HasTag(name="TELEPATHIC")
            & ~OwnerHasTech(name="LRN_PSY_DOM")
            & ContainedBy(
                System
                & Contains(
                    Planet()
                    & OwnedBy(empire=Source.Owner)
                    &
                    # @content_tag{TELEPATHIC} On owned planet with focus FOCUS_DOMINATION and this tag, chance of taking ownership of suitable ships
                    HasTag(name="TELEPATHIC")
                    & Focus(type=["FOCUS_DOMINATION"])
                )
            ),
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_PSY_DOM",
                    label="EFFECT_PSY_DOM_LABEL",
                    icon="icons/sitrep/ship_produced.png",
                    parameters={"empire": Target.Owner, "ship": Target.ID},
                    empire=Target.Owner,
                ),
                GenerateSitRepMessage(
                    message="EFFECT_PSY_DOM",
                    label="EFFECT_PSY_DOM_LABEL",
                    icon="icons/sitrep/ship_produced.png",
                    parameters={"empire": Target.Owner, "ship": Target.ID},
                    empire=Source.Owner,
                ),
                SetOwner(empire=Source.Owner),
            ],
        )
    ],
    graphic="icons/tech/psychogenic_domination.png",
)



from focs._tech import *
from techs.ship_weapons.ship_weapons import WEAPON_BASE_EFFECTS

Tech(
    name="SHP_ROOT_AGGRESSION",
    description="SHP_ROOT_AGGRESSION_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_SHIP_WEAPONS_CATEGORY"],
    unlock=[
        Item(type=UnlockShipPart, name="SR_WEAPON_1_1"),
        Item(type=UnlockShipPart, name="GT_TROOP_POD"),
        Item(type=UnlockShipPart, name="SR_WEAPON_0_1"),
    ],
    effectsgroups=[
        # TODO move at least the flak effect to the part definition
        *WEAPON_BASE_EFFECTS("SR_WEAPON_0_1"),
        *WEAPON_BASE_EFFECTS("SR_WEAPON_1_1"),
    ],
    graphic="icons/tech/planetary_colonialism.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_DOMESTIC_MONSTER",
    description="SHP_DOMESTIC_MONSTER_DESC",
    short_description="TAME_MONSTER_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=90 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=[
        "GRO_MEGA_ECO",
        "LRN_TRANSLING_THT",
        "GRO_GENETIC_MED",
    ],
    graphic="icons/monsters/kraken-5.png",
)



from freeorion_tools import assertion_fails


def test_does_fail():
    assert assertion_fails(False)


def test_does_not_fail():
    assert not assertion_fails(True)


def test_message_argument():
    assert assertion_fails(False, "")
    assert assertion_fails(False, "Some message")
    assert not assertion_fails(True, "")
    assert not assertion_fails(True, "Some message")



from collections.abc import Iterator, Mapping
from typing import TypeVar

KEY = TypeVar("KEY")
VAL = TypeVar("VAL")

_mapper_registry = []


class Mapper(Mapping[KEY, VAL]):
    def __init__(self, name: str, mapping: dict):
        self._dict = mapping
        self._used = set()
        self.name = name
        _mapper_registry.append(self)

    def __getitem__(self, key: KEY) -> VAL:
        self._used.add(key)
        return self._dict[key]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterator[KEY]:
        return iter(self._dict)

    def unused_items(self) -> set[KEY]:
        return set(self._dict) - self._used


def report_unused_mapping():
    for mapper in _mapper_registry:
        unused_items = mapper.unused_items()
        if unused_items:
            print(f"Mapper '{mapper.name}' has unused items for this stub: {sorted(unused_items)}")



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_EVERYTHING",
    description="LRN_EVERYTHING_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_LEARNING_CATEGORY", "THEORY"],
    prerequisites=["LRN_GRAVITONICS"],
    unlock=Item(type=UnlockBuilding, name="BLD_FIELD_REPELLOR"),
    graphic="icons/tech/the_theory_of_everything.png",
)



# EffectsGroups for System Defense Mines
# arg1 total damage
# arg2 priority
# arg3 scope of systems effected: SOURCE, EMPIRE
#
# This macro handles multiple entries, where only the earliest applicable effects a target.
#   arg2/priority: the value should be unique from any other uses of EG_SYSTEM_MINES
#   arg3/scope: applies to either the source system only, or to all systems with an empire owned planet.
# Initial priority order of entries (earliest first):
#   SYS_DEF_MINES_3
#   SYS_DEF_MINES_2
#   FORTRESS_SPECIAL
#   SYS_DEF_MINES_1

from focs._effects import (
    ContainedBy,
    Contains,
    Destroy,
    EffectsGroup,
    EnemyOf,
    Fleet,
    GenerateSitRepMessage,
    HasTag,
    InSystem,
    OwnedBy,
    Planet,
    SetStructure,
    Ship,
    Source,
    Structure,
    System,
    Target,
    Unowned,
    Value,
    VisibleToEmpire,
)


def EG_SYSTEM_MINES(total_damage, priority, scope: str):
    return [
        # Empire owned planet, damage enemy ship; sitreps issued below
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned)
            & Structure(low=total_damage + 0.0001),
            activation=~Unowned,
            stackinggroup="DEF_MINES_DAMAGE_STACK",
            priority=priority,
            effects=[SetStructure(value=Value - total_damage)],
        ),
        # Unowned planet, damage non-friendly ship; sitreps issued below
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (~Unowned | (Unowned & (~HasTag(name="UNOWNED_FRIENDLY"))))
            & Structure(low=total_damage + 0.0001),
            activation=Unowned,
            stackinggroup="DEF_MINES_DAMAGE_STACK",
            priority=priority,
            effects=[SetStructure(value=Value - total_damage)],
        ),
        # Empire owned planet, destroy enemy ship; sitreps issued below
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned)
            & Structure(high=total_damage),
            activation=~Unowned,
            stackinggroup="DEF_MINES_DAMAGE_STACK",
            priority=priority,
            effects=[Destroy],
        ),
        # Unowned planet, destroy non-friendly ship; sitreps issued below
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (~Unowned | (Unowned & (~HasTag(name="UNOWNED_FRIENDLY"))))
            & Structure(high=total_damage),
            activation=Unowned,
            stackinggroup="DEF_MINES_DAMAGE_STACK",
            priority=priority,
            effects=[Destroy],
        ),
        # Sitreps
        # Planet owner sitrep - visible enemy fleet damaged
        EffectsGroup(
            scope=Fleet
            & SYSTEM_MINES_SCOPE(scope)
            & OwnedBy(affiliation=EnemyOf, empire=Source.Owner)
            & VisibleToEmpire(empire=Source.Owner),
            activation=~Unowned,
            stackinggroup="DEF_MINES_SITREP_PLANET_VISIBLE_ENEMY_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES",
                    label="EFFECT_MINES_LABEL",
                    icon="icons/sitrep/combat_damage.png",
                    parameters={
                        "empire": Target.Owner,
                        "fleet": Target.ID,
                        "rawtext": total_damage,
                        "system": Target.SystemID,
                    },
                    empire=Source.Owner,
                )
            ],
        ),
        # Planet owner sitrep - visible unowned ship damaged
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & Unowned
            & VisibleToEmpire(empire=Source.Owner)
            & Structure(low=total_damage + 0.0001),
            activation=~Unowned,
            stackinggroup="DEF_MINES_SITREP_PLANET_VISIBLE_UNOWNED_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES_SINGLE_SHIP",
                    label="EFFECT_MINES_LABEL",
                    icon="icons/sitrep/combat_damage.png",
                    parameters={
                        "ship": Target.ID,
                        "shipdesign": Target.DesignID,
                        "rawtext": total_damage,
                        "system": Target.SystemID,
                    },
                    empire=Source.Owner,
                )
            ],
        ),
        # Planet owner sitrep - visible unowned ship destroyed
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & Unowned
            & VisibleToEmpire(empire=Source.Owner)
            & Structure(high=total_damage),
            activation=~Unowned,
            stackinggroup="DEF_MINES_SITREP_PLANET_VISIBLE_UNOWNED_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES_UNOWNED_DESTROYED",
                    label="EFFECT_MINES_SHIP_DESTROYED_LABEL",
                    icon="icons/sitrep/combat_damage.png",
                    parameters={"ship": Target.ID, "shipdesign": Target.DesignID, "system": Target.SystemID},
                    empire=Source.Owner,
                )
            ],
        ),
        # Planet owner sitrep - visible enemy ship destroyed
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & OwnedBy(affiliation=EnemyOf, empire=Source.Owner)
            & Structure(high=total_damage)
            & VisibleToEmpire(empire=Source.Owner),
            activation=~Unowned,
            stackinggroup="DEF_MINES_SITREP_PLANET_VISIBLE_ENEMY_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES_SHIP_DESTROYED",
                    label="EFFECT_MINES_SHIP_DESTROYED_LABEL",
                    icon="icons/sitrep/combat_destroyed.png",
                    parameters={"empire": Target.Owner, "ship": Target.ID, "system": Target.SystemID},
                    empire=Source.Owner,
                )
            ],
        ),
        # Planet owner sitrep - non-visible ship damaged or destroyed
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned)
            & (~VisibleToEmpire(empire=Source.Owner)),
            activation=~Unowned,
            stackinggroup="DEF_MINES_SITREP_PLANET_INVISIBLE_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES_UNKNOWN",
                    label="EFFECT_MINES_UNKNOWN_LABEL",
                    icon="icons/sitrep/combat_damage.png",
                    parameters={"system": Target.SystemID},
                    empire=Source.Owner,
                )
            ],
        ),
        # Fleet owner sitrep - damaged fleet
        EffectsGroup(
            scope=Fleet & SYSTEM_MINES_SCOPE(scope) & (~Unowned) & OwnedBy(affiliation=EnemyOf, empire=Source.Owner),
            stackinggroup="DEF_MINES_SITREP_FLEET_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES",
                    label="EFFECT_MINES_LABEL",
                    icon="icons/sitrep/combat_damage.png",
                    parameters={
                        "empire": Target.Owner,
                        "fleet": Target.ID,
                        "rawtext": total_damage,
                        "system": Target.SystemID,
                    },
                    empire=Target.Owner,
                )
            ],
        ),
        # Ship owner sitrep - destroyed ship
        EffectsGroup(
            scope=Ship
            & SYSTEM_MINES_SCOPE(scope)
            & (~Unowned)
            & OwnedBy(affiliation=EnemyOf, empire=Source.Owner)
            & Structure(high=total_damage),
            stackinggroup="DEF_MINES_SITREP_SHIP_STACK",
            priority=priority,
            effects=[
                GenerateSitRepMessage(
                    message="EFFECT_MINES_SHIP_DESTROYED",
                    label="EFFECT_MINES_SHIP_DESTROYED_LABEL",
                    icon="icons/sitrep/combat_destroyed.png",
                    parameters={"empire": Target.Owner, "ship": Target.ID, "system": Target.SystemID},
                    empire=Target.Owner,
                )
            ],
        ),
    ]


def SYSTEM_MINES_SCOPE(scope: str):
    if scope == "SOURCE":
        return InSystem(id=Source.SystemID)
    elif scope == "EMPIRE":
        return ContainedBy(System & Contains(Planet() & OwnedBy(empire=Source.Owner)))
    else:
        raise Exception("Unknown mines scope %s" % scope)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_BIOTERM",
    description="SHP_BIOTERM_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_BOMBARD_WEAPON_TECHS"],
    prerequisites=["SHP_DEATH_SPORE", "GRO_BIOTERROR"],
    unlock=Item(type=UnlockShipPart, name="SP_BIOTERM"),
    graphic="icons/ship_parts/bioterm.png",
)



import pytest

from stub_generator.stub_generator.collection_classes import is_collection_type, make_type


@pytest.mark.parametrize(
    ("string, hint"),
    (
        # Base types
        ("Int", "int"),
        ("Float", "float"),
        ("Double", "float"),
        ("String", "str"),
        ("Bool", "bool"),
        # Simple collections
        ("IntPair", "tuple[int, int]"),
        ("IntSet", "Set[int]"),
        ("IntVec", "Sequence[int]"),
        # Embedded collections
        ("IntSetSet", "Set[Set[int]]"),
        # mapping
        ("StringIntMap", "Mapping[str, int]"),
        ("IntSetFloatMap", "Mapping[Set[int], float]"),
        ("IntMeterTypeAccountingInfoVecMapMap", "Mapping[int, Mapping[meterType, Sequence[AccountingInfo]]]"),
        # Not matchable
        ("Hello", "Hello"),
        ("HelloVec", "HelloVec"),
        # Empty
        ("", ""),
        (None, ""),
        # Underscores are ignored
        ("_I_n_t_", "int"),
        ("Int_Set", "Set[int]"),
        ("Int_MeterType_AccountingInfoVecMapMap", "Mapping[int, Mapping[meterType, Sequence[AccountingInfo]]]"),
    ),
)
def test_make_type(string, hint):
    assert make_type(string) == hint


@pytest.mark.parametrize(
    "type_string",
    ["IntSet", "IntVec", "IntIntMap", "AccountingInfoVec", "IntDoubleMap", "MeterTypeAccountingInfoVecPair"],
)
def test_collection_type_is_true(type_string):
    assert is_collection_type(type_string) is True


@pytest.mark.parametrize("type_string", ["Int", "Hello", "universe"])
def test_collection_type_is_false(type_string):
    assert is_collection_type(type_string) is False



from focs._effects import (
    AsteroidsType,
    Barren,
    Contains,
    Destroy,
    EffectsGroup,
    GasGiantType,
    GenerateSitRepMessage,
    IsBuilding,
    IsSource,
    Object,
    OwnedBy,
    Planet,
    Population,
    SetPlanetType,
    Source,
    Target,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_ARTIFICIAL_PLANET_EXCLUSION

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_ART_PLANET",
    description="BLD_ART_PLANET_DESC",
    buildcost=200 * Target.HabitableSize * BUILDING_COST_MULTIPLIER,
    buildtime=8,
    location=(
        Planet()
        & ~Contains(IsBuilding(name=["BLD_ART_PLANET"]))
        & ~Contains(IsBuilding(name=["BLD_ART_FACTORY_PLANET"]))
        & ~Contains(IsBuilding(name=["BLD_ART_PARADISE_PLANET"]))
        & OwnedBy(empire=Source.Owner)
        & Planet(type=[AsteroidsType, GasGiantType])
        & Population(high=0)
    ),
    enqueuelocation=ENQUEUE_ARTIFICIAL_PLANET_EXCLUSION,
    effectsgroups=[
        EffectsGroup(
            scope=Object(id=Source.PlanetID) & Planet(),
            effects=[
                SetPlanetType(type=Barren),
                GenerateSitRepMessage(
                    message="EFFECT_ART_PLANET",
                    label="EFFECT_ART_PLANET_LABEL",
                    icon="icons/specials_large/tidal_lock.png",
                    parameters={"planet": Target.ID},
                    empire=Source.Owner,
                ),
            ],
        ),
        EffectsGroup(scope=IsSource, effects=Destroy),
    ],
    icon="icons/specials_huge/tidal_lock.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_BIOADAPTIVE_SPEC",
    description="SHP_BIOADAPTIVE_SPEC_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=216 * TECH_COST_MULTIPLIER,
    researchturns=12,
    tags=["PEDIA_ORGANIC_HULL_TECHS"],
    prerequisites=["SHP_MONOCELL_EXP", "GRO_TRANSORG_SENT"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_BIOADAPTIVE"),
    ],
    graphic="icons/ship_hulls/bio_adaptive_hull_small.png",
)



from empire.survey_lock import survey_universe_lock
from freeorion_tools.caching import cache_for_current_turn


class _ColonyStatus:
    def __init__(self):
        self.colonies_under_attack = False
        self.colonies_under_threat = False


@survey_universe_lock
def colonies_is_under_attack() -> bool:
    return bool(_get_colony_status().colonies_under_attack)


@survey_universe_lock
def colonies_is_under_treat() -> bool:
    return bool(_get_colony_status().colonies_under_threat)


@cache_for_current_turn
def _get_colony_status() -> _ColonyStatus:
    return _ColonyStatus()


def set_colonies_is_under_treat():
    _get_colony_status().colonies_under_threat = True


def set_colonies_is_under_attack():
    _get_colony_status().colonies_under_attack = True



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    IsSource,
    LocalCandidate,
    MinOf,
    Planet,
    SetMaxSupply,
    SetSupply,
    Target,
    Value,
)
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

STANDARD_SUPPLY_GROWTH = EffectsGroup(  # increase 1 per turn, up to max
    scope=IsSource,
    activation=Planet()
    & (LocalCandidate.LastTurnConquered < CurrentTurn)
    & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
    priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    effects=SetSupply(value=MinOf(float, Value(Target.MaxSupply), Value + 1)),
)


VERY_BAD_SUPPLY = [
    EffectsGroup(
        description="VERY_BAD_SUPPLY_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="VERY_BAD_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value - 1),
    ),
    STANDARD_SUPPLY_GROWTH,
]


BAD_SUPPLY = [
    EffectsGroup(
        description="BAD_SUPPLY_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="BAD_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value + 0),
    ),
    STANDARD_SUPPLY_GROWTH,
]


AVERAGE_SUPPLY = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet(),
        accountinglabel="AVERAGE_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value + 1),
    ),
    STANDARD_SUPPLY_GROWTH,
]

GREAT_SUPPLY = [
    EffectsGroup(
        description="GREAT_SUPPLY_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GREAT_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value + 2),
    ),
    STANDARD_SUPPLY_GROWTH,
]

ULTIMATE_SUPPLY = [
    EffectsGroup(
        description="ULTIMATE_SUPPLY_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="ULTIMATE_SUPPLY_LABEL",
        effects=SetMaxSupply(value=Value + 3),
    ),
    STANDARD_SUPPLY_GROWTH,
]



from collections import Counter
from typing import Optional

from CombatRatingsAI._fleet_combat_stats import get_fleet_combat_stats
from CombatRatingsAI._ship_combat_stats import AttackCount, AttackDamage, ShipCombatStats, get_ship_combat_stats
from common.fo_typing import FleetId, ShipId, SpeciesName
from EnumsAI import MissionType
from freeorion_tools import get_species_ship_shields
from freeorion_tools.caching import cache_for_current_turn


@cache_for_current_turn
def get_empire_standard_military_ship_stats() -> ShipCombatStats:
    """Get the current empire standard military ship stats, i.e. the most common ship type within the empire.

    :return: Stats of most common military ship in the empire
    """

    # avoid circular imports
    from FleetUtilsAI import get_empire_fleet_ids_by_role

    stats_dict = Counter()
    for fleet_id in get_empire_fleet_ids_by_role(MissionType.MILITARY):
        ship_stats = get_fleet_combat_stats(fleet_id, max_stats=True).get_ship_combat_stats()
        stats_dict.update(ship_stats)

    most_commons = stats_dict.most_common(1)
    if most_commons:
        return most_commons[0][0]
    else:
        return default_ship_stats()


def default_ship_stats() -> ShipCombatStats:
    """Return some ship stats to assume if no other intel is available.

    :return: Some weak standard ship
    """
    return ShipCombatStats(
        attacks={AttackDamage(6.0): AttackCount(1)},
        structure=15,
        shields=0,
        fighter_capacity=0,
        fighter_launch_rate=0,
        fighter_damage=0,
        flak_shots=0,
        has_interceptors=False,
        damage_vs_planets=0,
        has_bomber=False,
    )


def species_shield_bonus(species: Optional[SpeciesName], shield_type: Optional[str]) -> float:
    skill = get_species_ship_shields(species)
    shield_class = {
        "SH_DEFENSE_GRID": 1,
        "SH_DEFLECTOR": 2,
        "SH_PLASMA": 3,
        "SH_MULTISPEC": 4,
        "SH_BLACK": 5,
    }
    # Ships without or with robotic shield get no bonus, see shields.macro
    return skill * shield_class.get(shield_type, 0)


def rating_needed(target: float, current: float = 0) -> float:
    """Estimate the needed rating to achieve target rating.

    :param target: Target rating to be reached
    :param current: Already existing rating
    :return: Estimated missing rating to reach target
    """
    if current >= target or target <= 0:
        return 0
    else:
        return target + current - 2 * (target * current) ** 0.5


def rating_difference(first_rating: float, second_rating: float) -> float:
    """Return the absolute nonlinear difference between ratings.

    :param first_rating: rating of a first force
    :param second_rating: rating of a second force
    :return: Estimated rating by which the greater force (nonlinearly) exceeds the lesser
    """

    return rating_needed(max(first_rating, second_rating), min(first_rating, second_rating))


def get_fleet_rating(fleet_id: FleetId, enemy_stats: ShipCombatStats = None) -> float:
    """Get rating for the fleet against specified enemy.

    :param fleet_id: fleet to be rated
    :param enemy_stats: enemy to be rated against
    """
    return get_fleet_combat_stats(fleet_id, max_stats=False).get_rating(enemy_stats)


def get_fleet_rating_against_planets(fleet_id: FleetId):
    return get_fleet_combat_stats(fleet_id, max_stats=False).get_rating_vs_planets()


def get_ship_rating(ship_id: ShipId, enemy_stats=None):
    return get_ship_combat_stats(ship_id=ship_id, max_stats=False).get_rating(enemy_stats)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import ASTEROIDAL_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.general import FAST_COLONIZATION
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import VERY_BAD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import VERY_BAD_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, NO_OFFENSE_TROOPS

Species(
    name="SP_CELESTEPHYTE",
    description="SP_CELESTEPHYTE_DESC",
    gameplay_description="SP_CELESTEPHYTE_GAMEPLAY_DESC",
    native=True,
    tags=[
        "PHOTOTROPHIC",
        "TELEPATHIC",
        "VERY_BAD_POPULATION",
        "BAD_HAPPINESS",
        "BAD_SUPPLY",
        "VERY_BAD_INDUSTRY",
        "NO_OFFENSE_TROOPS",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_GROWTH",
    likes=[
        "BLD_LIGHTHOUSE",  # if there were a greenhouse they would like that too
        "FOCUS_GROWTH",  # what else would a plant want?
        "PLC_POPULATION",  # grow, grow, grow your load, gently down the ..
    ],
    dislikes=[  # dislikes anything disturbing their fragile habitat or irritating their telepathic minds
        "PLC_NO_GROWTH",  # is a plant, they like to grow
        "BLD_SHIPYARD_AST",  # disturbs the rocks
        "BLD_SHIPYARD_AST_REF",  # same
        "JUGGERNAUT_NEST_SPECIAL",  # compete for rocks
        "ECCENTRIC_ORBIT_SPECIAL",  # messes up all those orbital resonances, quite inharmonious
        "FLD_METEOR_BLIZZARD",  # same
        "FLD_VOID_RIFT",  # same
    ],
    effectsgroups=[
        *VERY_BAD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *VERY_BAD_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_CELESTEPHYTE"),
        *BAD_SUPPLY,
        *NO_OFFENSE_TROOPS,
        *BAD_DEFENSE_TROOPS,
        FAST_COLONIZATION,
        # not for description
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        NARROW_EP,
    ],
    environments=ASTEROIDAL_NARROW_EP,
    graphic="icons/species/celestephyte.png",
)



from focs._effects import (
    Capital,
    EffectsGroup,
    Focus,
    HasSpecies,
    OwnedBy,
    Planet,
    SetMaxStockpile,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import STOCKPILE_PER_POP, TECH_COST_MULTIPLIER

Tech(
    name="PRO_GENERIC_SUPPLIES",
    description="PRO_GENERIC_SUPPLIES_DESC",
    short_description="IMPERIAL_STOCKPILE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=40 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=[
        "PRO_PREDICTIVE_STOCKPILING",
        "CON_ASYMP_MATS",
        "PRO_ROBOTIC_PROD",
    ],
    effectsgroups=[
        EffectsGroup(
            scope=Capital & OwnedBy(empire=Source.Owner),
            effects=SetMaxStockpile(value=Value + 2, accountinglabel="GENERIC_SUPPLIES_FIXED_BONUS_LABEL"),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & HasSpecies(),
            effects=SetMaxStockpile(
                value=(Value + 0.5 * Target.Population * STOCKPILE_PER_POP),
                accountinglabel="GENERIC_SUPPLIES_POPULATION_BONUS_LABEL",
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Focus(type=["FOCUS_STOCKPILE"]),
            effects=SetMaxStockpile(value=Value + 3, accountinglabel="GENERIC_SUPPLIES_FOCUS_BONUS_LABEL"),
        ),
    ],
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BARREN_BROAD_EP, BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GREAT_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import GOOD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS

Species(
    name="SP_CRAY",
    description="SP_CRAY_DESC",
    gameplay_description="SP_CRAY_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "ARTISTIC",
        "BAD_INDUSTRY",
        "GOOD_RESEARCH",
        "GOOD_INFLUENCE",
        "GREAT_HAPPINESS",
        "AVERAGE_SUPPLY",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "SUCCULENT_BARNACLES_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "MONOPOLE_SPECIAL",
        "CRYSTALS_SPECIAL",
        "FORTRESS_SPECIAL",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_DIVINE_AUTHORITY",
        "PLC_DIVERSITY",
        "PLC_CONFEDERATION",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_ALLIED_REPAIR",
    ],
    dislikes=[
        "BLD_HYPER_DAM",
        "BLD_GAS_GIANT_GEN",
        "BLD_SCANNING_FACILITY",
        "BLD_SOL_ORB_GEN",
        "BLD_INDUSTRY_CENTER",
        "KRAKEN_NEST_SPECIAL",
        "WORLDTREE_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PLC_CONFORMANCE",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_CENTRALIZATION",
        "PLC_TERROR_SUPPRESSION",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *GOOD_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GREAT_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_CRAY"),
        *AVERAGE_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        BROAD_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=BARREN_BROAD_EP,
    graphic="icons/species/cray.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_FRC_ENRG_COMP",
    description="SHP_FRC_ENRG_COMP_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=126 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_ENERGY_HULL_TECHS"],
    prerequisites=["LRN_FORCE_FIELD"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_COMPRESSED_ENERGY"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_ENRG_COMP"),
        Item(type=UnlockPolicy, name="PLC_CHARGE"),
    ],
    graphic="icons/ship_hulls/compressed_energy_hull_small.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ASTEROID_REFORM",
    description="SHP_ASTEROID_REFORM_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=180 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["SHP_ASTEROID_HULLS"],
    unlock=[
        # Item type = ShipHull name = "SH_AGREGATE_ASTEROID"
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_AST_REF"),
        Item(type=UnlockShipPart, name="AR_ROCK_PLATE"),
    ],
    graphic="icons/ship_parts/rock_plating.png",
)



from focs._effects import (
    AnyEmpire,
    Contains,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    GameRule,
    IsSource,
    IsTarget,
    NamedIntegerLookup,
    NoEffect,
    OwnedBy,
    Planet,
    StatisticIf,
    System,
    WithinStarlaneJumps,
)

MIN_RECOLONIZING_SIZE = 3

MIN_RECOLONIZING_HAPPINESS = 5

IMPOSSIBLY_LARGE_TURN = 2**15


def DESCRIPTION_EFFECTSGROUP_MACRO(desc: str):
    return EffectsGroup(description=desc, scope=IsSource, activation=None, effects=NoEffect)


FIGHTER_DAMAGE_FACTOR = GameRule(type=float, name="RULE_FIGHTER_DAMAGE_FACTOR")

PLANET_DEFENSE_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

PLANET_SHIELD_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SHIP_WEAPON_DAMAGE_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

SHIP_SHIELD_FACTOR = GameRule(type=float, name="RULE_SHIP_WEAPON_DAMAGE_FACTOR")

SHIP_STRUCTURE_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SYSTEM_MINES_DAMAGE_FACTOR = GameRule(type=float, name="RULE_SHIP_STRUCTURE_FACTOR")

SUPPLY_DISCONNECTED_INFLUENCE_MALUS = 1

# empire id used for unowned planets/ships - as defined in Universe.cpp(?)
UNOWNED_EMPIRE_ID = -1  # type: ignore[assignment]


MINIMUM_DISTANCE_EMPIRE_CHECK = ~WithinStarlaneJumps(
    jumps=NamedIntegerLookup(name="MIN_MONSTER_DISTANCE"),
    condition=System & Contains(Planet() & OwnedBy(affiliation=AnyEmpire)),
)

GROWTH_RATE_FACTOR = (
    0.1
    * (
        1 - StatisticIf(float, condition=IsTarget & EmpireHasAdoptedPolicy(name="PLC_NO_GROWTH"))
    )  # no growth with no-growth policy
    * (
        1 + 0.5 * StatisticIf(float, condition=IsTarget & EmpireHasAdoptedPolicy(name="PLC_POPULATION"))
    )  # +50% growth with population policy
)



from focs._effects import Capital, EffectsGroup, GiveEmpirePolicy, IsSource, Planet, Unowned
from focs._species import *
from species.species_macros.detection import GOOD_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, RADIATED_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GREAT_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import GOOD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.telepathic import TELEPATHIC_DETECTION
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS
from species.species_macros.xenophobic import XENOPHOBIC_OTHER, XENOPHOBIC_SELF

Species(
    name="SP_TRITH",
    description="SP_TRITH_DESC",
    gameplay_description="SP_TRITH_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "SELF_SUSTAINING",
        "TELEPATHIC",
        "XENOPHOBIC",
        "GREAT_INFLUENCE",
        "GOOD_SHIP_SHIELDS",
        "BAD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "GOOD_DETECTION",
        "GOOD_STEALTH",
        "PEDIA_SELF_SUSTAINING_SPECIES_CLASS",
        "PEDIA_XENOPHOBIC_SPECIES_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "SPARK_FOSSILS_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "MONOPOLE_SPECIAL",
        "MINERALS_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "PLC_CHECKPOINTS",
        "PLC_RACIAL_PURITY",
        "PLC_CONFEDERATION",
        "PLC_TRAFFIC_CONTROL",
        "PLC_DREAM_RECURSION",
        "PLC_ISOLATION",
        "PLC_DIVINE_AUTHORITY",
    ],
    dislikes=[
        # "BLD_CLONING_CENTER"       // disabled content,
        "BLD_COLLECTIVE_NET",
        "BLD_GENOME_BANK",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_XENORESURRECTION_LAB",
        "FOCUS_INDUSTRY",
        "FORTRESS_SPECIAL",
        "PANOPTICON_SPECIAL",
        "PLC_ALLIED_REPAIR",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_DIVERSITY",
        "PLC_EXPLORATION",
        "SPICE_SPECIAL",
        "WORLDTREE_SPECIAL",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GREAT_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_TRITH"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *GOOD_DETECTION,
        *GOOD_STEALTH,
        *XENOPHOBIC_SELF,
        XENOPHOBIC_OTHER("TRITH"),
        *TELEPATHIC_DETECTION(5),
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        BROAD_EP,
        *GOOD_SHIP_SHIELDS,
        EffectsGroup(
            scope=IsSource, activation=Planet() & ~Unowned & Capital, effects=GiveEmpirePolicy(name="PLC_RACIAL_PURITY")
        ),
    ],
    environments=RADIATED_BROAD_EP,
    graphic="icons/species/trith.png",
)



import os
from collections.abc import Generator


def get_code_location(generator: Generator) -> str:
    if generator.gi_yieldfrom:
        return get_code_location(generator.gi_yieldfrom)
    return f"{os.path.basename(generator.gi_code.co_filename)}:{generator.gi_frame.f_lineno}"


def get_item_with_location(generator: Generator) -> Generator:
    for x in generator:
        yield get_code_location(generator), x


def handle_interfaces_mismatch():
    """
    Handle cases when AI and universe generation interfaces mismatch.
    """

    try:
        import freeorion as fo

        universe = fo.get_universe()
        empire_of_first_ai = fo.get_empire(2)
        galaxy_data = fo.get_galaxy_setup_data()
        return fo, universe, empire_of_first_ai, galaxy_data
    except ImportError:
        pass

    try:
        import freeOrionAIInterface as fo

        universe = fo.getUniverse()
        empire_of_first_ai = fo.getEmpire(2)
        galaxy_data = fo.getGalaxySetupData()
        return fo, universe, empire_of_first_ai, galaxy_data

    except ImportError:
        pass


def get_common_instances() -> Generator:
    """
    Get instances for objects that are the same to AI and universe generation interfaces.
    """
    fo, universe, empire_of_first_ai, galaxy_data = handle_interfaces_mismatch()

    yield universe
    yield galaxy_data

    planet = universe.getPlanet(empire_of_first_ai.capitalID)
    yield planet

    yield universe.getSystem(planet.systemID)

    tech = fo.getTech("SHP_WEAPON_2_1")
    yield tech
    yield tech.unlockedItems[0]

    yield fo.getGameRules()
    ship_hull = fo.getShipHull("SH_XENTRONIUM")
    yield ship_hull

    yield fo.getSpecies("SP_CRAY")

    fleets_int_vector = universe.fleetIDs

    fleet = universe.getFleet(list(fleets_int_vector)[0])
    yield fleet

    fields_ids = universe.fieldIDs
    field = universe.getField(fields_ids[0])
    yield field

    yield fo.getFieldType("FLD_ION_STORM")
    yield fo.getBuildingType("BLD_SHIPYARD_BASE")

    yield fo.getShipPart("SR_WEAPON_1_1")
    yield fo.getSpecial("MODERATE_TECH_NATIVES_SPECIAL")
    yield fo.getShipHull("SH_XENTRONIUM")

    ship = universe.getShip(list(universe.shipIDs)[0])
    yield ship

    design = fo.getShipDesign(ship.designID)
    yield design

    yield fo.diplomaticMessage(1, 2, fo.diplomaticMessageType.acceptPeaceProposal)
    yield empire_of_first_ai
    yield empire_of_first_ai.productionQueue
    yield empire_of_first_ai.researchQueue

    planet = universe.getPlanet(empire_of_first_ai.capitalID)
    yield planet

    meter = planet.getMeter(fo.meterType.population)
    yield meter

    building = list(planet.buildingIDs)[0]
    yield universe.getBuilding(building)

    yield fo.getPolicy("PLC_LIBERTY")


common_classes_to_exclude = {
    "popCenter",  # parent class, it's not possible to get instance
    "resourceCenter",  # parent class, it's not possible to get instance
    "diplomaticStatusUpdate",  # this item is not used in generate orders/universe
    "universeObject",  # parent class, it's not possible to get instance
}

classes_to_exclude_from_universe = {
    "productionQueueElement",  # not applicable
    "researchQueueElement",  # not applicable
    *common_classes_to_exclude,
}

classes_to_exclude_from_ai = {
    *common_classes_to_exclude,
}



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SPY_STEALTH_PART_3",
    description="SPY_STEALTH_PART_3_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=350 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_STEALTH_PART_TECHS"],
    prerequisites=["SPY_DETECT_4"],
    unlock=Item(type=UnlockShipPart, name="ST_CLOAK_3"),
    graphic="icons/ship_parts/cloak-3.png",
)



from typing import TypeVar

class _GameRuleDefinition: ...

R = TypeVar("R", int, float, str)

def GameRule(
    *,
    type: type[R],
    name: str,
    description="",
    category="",
    default: R = ...,
    min: float = 0.0,
    max: float = 0.0,
    allowed: list[str] = ...,
) -> _GameRuleDefinition: ...



from focs._effects import (
    AsteroidsType,
    ContainedBy,
    Contains,
    EffectsGroup,
    Focus,
    Happiness,
    IsSource,
    NamedReal,
    OwnedBy,
    Planet,
    SetTargetIndustry,
    Source,
    System,
    Target,
    Value,
)
from focs._species import *
from macros.base_prod import INDUSTRY_PER_POP
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import RADIATED_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import VERY_BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import GREAT_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import BAD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_KILANDOW",
    description="SP_KILANDOW_DESC",
    gameplay_description="SP_KILANDOW_GAMEPLAY_DESC",
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "PHOTOTROPHIC",
        "GREAT_RESEARCH",
        "VERY_BAD_INFLUENCE",
        "BAD_STEALTH",
        "CTRL_EXTINCT",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "FRACTAL_GEODES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "CRYSTALS_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_EXPLORATION",
        "PLC_LIBERTY",
        "PLC_NATIVE_APPROPRIATION",
        "PLC_TECHNOCRACY",
    ],
    dislikes=[
        # "BLD_CLONING_CENTER"       // disabled content,
        "BLD_SCANNING_FACILITY",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_CON_NANOROBO",
        "BLD_SHIPYARD_ORG_ORB_INC",
        "BLD_SPACE_ELEVATOR",
        "ELERIUM_SPECIAL",
        "MONOPOLE_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_CHECKPOINTS",
        "PLC_DIVINE_AUTHORITY",
        "PLC_MARTIAL_LAW",
        "RESONANT_MOON_SPECIAL",
        "WORLDTREE_SPECIAL",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *GREAT_RESEARCH,
        *VERY_BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_KILANDOW"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *BAD_STEALTH,
        EffectsGroup(
            description="GREAT_ASTEROID_INDUSTRY_DESC",
            scope=IsSource
            & Focus(type=["FOCUS_INDUSTRY"])
            & Happiness(low=0)
            & ContainedBy(System & Contains(Planet() & Planet(type=[AsteroidsType]) & OwnedBy(empire=Source.Owner))),
            effects=SetTargetIndustry(
                value=Value
                + Target.Population
                * NamedReal(name="GREAT_ASTEROID_INDUSTRY_TARGET_INDUSTRY_PERPOP", value=1.0 * INDUSTRY_PER_POP)
            ),
        ),
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=RADIATED_STANDARD_EP,
    graphic="icons/species/kilandow.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import SWAMP_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
)
from species.species_macros.general import FAST_COLONIZATION
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import NO_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_PHINNERT",
    description="SP_PHINNERT_DESC",
    gameplay_description="SP_PHINNERT_GAMEPLAY_DESC",
    native=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "NO_RESEARCH",
        "GOOD_INFLUENCE",
        "GOOD_HAPPINESS",
        "BAD_SUPPLY",
        "PRIMITIVE",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "SUCCULENT_BARNACLES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "MINERALS_SPECIAL",
        "CRYSTALS_SPECIAL",
        "PANOPTICON_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_COLONIZATION",
        "PLC_POPULATION",
        "PLC_STOCKPILE_LIQUIDATION",
    ],
    dislikes=[
        "BLD_CULTURE_ARCHIVES",
        "BLD_CULTURE_LIBRARY",
        "BLD_REGIONAL_ADMIN",
        "BLD_MILITARY_COMMAND",
        "BLD_MEGALITH",
        "BLD_GENOME_BANK",
        "BLD_BLACK_HOLE_POW_GEN",
        "BLD_GAS_GIANT_GEN",
        "SPICE_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PLC_MARINE_RECRUITMENT",
        "PLC_NO_GROWTH",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        NO_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_PHINNERT"),
        *BAD_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        FAST_COLONIZATION,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=SWAMP_STANDARD_EP,
    graphic="icons/species/phinnert.png",
)



from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()

import freeorion as fo
import random

import universe_statistics
from common.handlers import init_handlers
from common.listeners import listener
from common.option_tools import parse_config
from empires import compile_home_system_list, setup_empire
from fields import generate_fields
from galaxy import calc_star_system_positions
from monsters import generate_monsters
from natives import generate_natives
from specials import distribute_specials
from starnames import name_star_systems
from starsystems import generate_systems, name_planets
from teams import place_teams
from universe_tables import MAX_JUMPS_BETWEEN_SYSTEMS, MAX_STARLANE_LENGTH
from util import error_list, int_hash, report_error, seed_rng

parse_config(fo.get_options_db_option_str("ai-config"), fo.get_user_config_dir())
init_handlers(fo.get_options_db_option_str("ai-config"), None)


class PyGalaxySetupData:
    """
    Class used to store and manage a copy of the galaxy setup data provided by the FreeOrion interface.
    This data can then be modified when needed during the universe generation process, without having to
    change the original data structure.
    """

    def __init__(self, galaxy_setup_data):
        self.seed = galaxy_setup_data.seed
        self.size = galaxy_setup_data.size
        self.shape = galaxy_setup_data.shape
        self.age = galaxy_setup_data.age
        self.starlane_frequency = galaxy_setup_data.starlaneFrequency
        self.planet_density = galaxy_setup_data.planetDensity
        self.specials_frequency = galaxy_setup_data.specialsFrequency
        self.monster_frequency = galaxy_setup_data.monsterFrequency
        self.native_frequency = galaxy_setup_data.nativeFrequency
        self.max_ai_aggression = galaxy_setup_data.maxAIAggression
        self.game_uid = galaxy_setup_data.gameUID

    def dump(self):
        print("Galaxy Setup Data:")
        print("...Seed:", self.seed)
        print("...Size:", self.size)
        print("...Shape:", self.shape)
        print("...Age:", self.age)
        print("...Starlane Frequency:", self.starlane_frequency)
        print("...Planet Density:", self.planet_density)
        print("...Specials Frequency:", self.specials_frequency)
        print("...Monster Frequency:", self.monster_frequency)
        print("...Native Frequency:", self.native_frequency)
        print("...Max AI Aggression:", self.max_ai_aggression)
        print("...Game UID:", self.game_uid)


def error_report():
    """
    Can be called from C++ to retrieve a list of errors that occurred during universe generation
    """
    return error_list


@listener
def create_universe(psd_map):  # noqa: C901
    """
    Main universe generation function invoked from C++ code.
    """
    print("Python Universe Generator")

    # fetch universe and player setup data
    gsd = PyGalaxySetupData(fo.get_galaxy_setup_data())
    gsd.dump()
    total_players = len(psd_map)

    # initialize RNG
    h = int_hash(gsd.seed.encode("utf-8"))
    print("Using hashed seed", h)
    seed_rng(h)
    seed_pool = [random.random() for _ in range(100)]
    print("Seed pool:", seed_pool)

    # make sure there are enough systems for the given number of players
    print("Universe creation requested with %d systems for %d players" % (gsd.size, total_players))
    min_size = total_players * 3
    if min_size > gsd.size:
        gsd.size = min_size
        print("Too few systems for the requested number of players, number of systems adjusted accordingly")
    print("Creating universe with %d systems for %d players" % (gsd.size, total_players))

    # calculate star system positions
    seed_rng(seed_pool.pop())
    system_positions = calc_star_system_positions(gsd)
    size = len(system_positions)
    print(gsd.shape, "Star system positions calculated, final number of systems:", size)

    # generate and populate systems
    seed_rng(seed_pool.pop())
    systems = generate_systems(system_positions, gsd)
    print(len(systems), "systems generated and populated")

    # generate Starlanes
    seed_rng(seed_pool.pop())
    fo.generate_starlanes(MAX_JUMPS_BETWEEN_SYSTEMS[gsd.starlane_frequency], MAX_STARLANE_LENGTH)
    print("Starlanes generated")

    print("Compile list of home systems...")
    seed_rng(seed_pool.pop())
    home_systems = compile_home_system_list(total_players, systems, gsd)
    if not home_systems:
        err_msg = "Python create_universe: couldn't get any home systems, ABORTING!"
        report_error(err_msg)
        raise Exception(err_msg)
    print("Home systems:", home_systems)

    teams = {}
    for psd in psd_map.values():
        if psd.starting_team >= 0:
            teams[psd.starting_team] = teams.get(psd.starting_team, 0) + 1
    teams = {k: v for k, v in teams.items() if v > 1}
    print("Teams: ", teams)

    seed_rng(seed_pool.pop())
    if teams:
        psds = list(psd_map.items())
        hs = list(home_systems)
        for home_system, team in place_teams(hs, systems, teams):
            hs.remove(home_system)
            psds_new = list()
            placed = False
            for empire, psd in psds:
                if placed or psd.starting_team != team:
                    psds_new.append((empire, psd))
                else:
                    placed = True
                    if not setup_empire(
                        empire, psd.empire_name, home_system, psd.starting_species, psd.player_name, gsd
                    ):
                        report_error("Python create_universe: couldn't set up empire for player %s" % psd.player_name)
            if not placed:
                report_error("Python create_universe: couldn't set up empire for team %d" % team)
            psds = psds_new
        # place leftovers
        for (empire, psd), home_system in zip(psds, hs):
            if not setup_empire(empire, psd.empire_name, home_system, psd.starting_species, psd.player_name, gsd):
                report_error("Python create_universe: couldn't set up empire for player %s" % psd.player_name)
    else:
        # set up empires for each player
        for empire, psd, home_system in zip(psd_map.keys(), psd_map.values(), home_systems):
            if not setup_empire(empire, psd.empire_name, home_system, psd.starting_species, psd.player_name, gsd):
                report_error("Python create_universe: couldn't set up empire for player %s" % psd.player_name)

    # assign names to all star systems and their planets
    # this needs to be done after all systems have been generated and empire home systems have been set, as
    # only after all that is finished star types as well as planet sizes and types are fixed, and the naming
    # process depends on that
    print("Assign star system names")
    seed_rng(seed_pool.pop())
    name_star_systems(systems)
    print("Set planet names")
    for system in systems:
        name_planets(system)

    print("Generating stationary fields in systems")
    seed_rng(seed_pool.pop())
    generate_fields(systems)

    print("Generating Natives")
    seed_rng(seed_pool.pop())
    generate_natives(gsd.native_frequency, systems, home_systems)

    print("Generating Space Monsters")
    seed_rng(seed_pool.pop())
    generate_monsters(gsd.monster_frequency, systems)

    print("Distributing Starting Specials")
    seed_rng(seed_pool.pop())
    distribute_specials(gsd.specials_frequency, fo.get_all_objects())

    # finally, write some statistics to the log file
    print("############################################################")
    print("##             Universe generation statistics             ##")
    print("############################################################")
    universe_statistics.log_planet_count_dist(systems)
    print("############################################################")
    universe_statistics.log_planet_type_summary(systems)
    print("############################################################")
    universe_statistics.log_species_summary(gsd.native_frequency)
    print("############################################################")
    universe_statistics.log_monsters_summary(gsd.monster_frequency)
    print("############################################################")
    universe_statistics.log_specials_summary()
    print("############################################################")
    universe_statistics.log_systems()
    universe_statistics.log_planets()

    if error_list:
        print("Python Universe Generator completed with errors")
        return False
    else:
        print("Python Universe Generator completed successfully")
        return True



from focs._effects import (
    EffectsGroup,
    NamedReal,
    OwnedBy,
    Planet,
    SetMaxShield,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import PLANET_SHIELD_FACTOR
from macros.priorities import DEFAULT_PRIORITY

Tech(
    name="LRN_FORCE_FIELD",
    description="LRN_FORCE_FIELD_DESC",
    short_description="SHIELD_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_NASCENT_AI"],
    unlock=Item(type=UnlockShipPart, name="SH_DEFENSE_GRID"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            stackinggroup="PLANET_SHIELDS_STACK_FF",
            priority=DEFAULT_PRIORITY,
            effects=SetMaxShield(
                value=Value + NamedReal(name="LRN_FORCE_FIELD_MAX_SHIELD_FLAT", value=10 * PLANET_SHIELD_FACTOR),
                accountinglabel="LRN_FORCE_FIELD",
            ),
        )
    ],
    graphic="icons/tech/forcefield_harmonics.png",
)



import freeorion as fo
import random
from itertools import cycle

# tuples of consonants and vowels for random name generation
consonants = ("b", "c", "d", "f", "g", "h", "j", "k", "l", "m", "n", "p", "q", "r", "s", "t", "v", "w", "x", "y", "z")
vowels = ("a", "e", "i", "o", "u")


_random_letter_generator = (random.choice(x) for x in cycle((consonants, vowels)))


def random_name(size):
    """
    Return random name of given size.

    It rotates between consonants and vowels.
    Rotation is global first letter depends on prev calls.
    """
    return "".join(next(_random_letter_generator) for _ in range(size)).capitalize()


def get_name_list(name_list):
    """
    Retrieves a list of names from the string tables.
    """
    return fo.userString(name_list).splitlines()



from focs._effects import AddSpecial, EffectsGroup, HasTag, OwnedBy, Planet, RemoveSpecial, Source, StatisticIf
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SPY_STEALTH_4",
    description="SPY_STEALTH_4_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=1200
    * TECH_COST_MULTIPLIER
    / (1 + StatisticIf(float, condition=OwnedBy(empire=Source.Owner) & HasTag(name="SNEAKY"))),
    researchturns=8,
    tags=["PEDIA_SPY_CATEGORY"],
    prerequisites=["SPY_STEALTH_3", "SPY_STEALTH_PART_3"],
    unlock=Item(type=UnlockShipPart, name="ST_CLOAK_4"),
    effectsgroups=[
        EffectsGroup(
            scope=OwnedBy(empire=Source.Owner) & Planet(),
            effects=[
                AddSpecial(name="VOID_SLAVE_SPECIAL"),
                RemoveSpecial(name="DIM_RIFT_SLAVE_SPECIAL"),
                RemoveSpecial(name="VOLCANIC_ASH_SLAVE_SPECIAL"),
                RemoveSpecial(name="CLOUD_COVER_SLAVE_SPECIAL"),
            ],
        )
    ],
    graphic="icons/specials_huge/void.png",
)



from focs._species import *
from species.species_macros.detection import BAD_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TOXIC_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.fuel import BAD_FUEL
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import BAD_STEALTH
from species.species_macros.stockpile import BAD_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS
from species.species_macros.weapons import ULTIMATE_WEAPONS

Species(
    name="SP_MISIORLA",
    description="SP_MISIORLA_DESC",
    gameplay_description="SP_MISIORLA_GAMEPLAY_DESC",
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "ARTISTIC",
        "BAD_INDUSTRY",
        "BAD_RESEARCH",
        "BAD_INFLUENCE",
        "BAD_STOCKPILE",
        "BAD_POPULATION",
        "GOOD_HAPPINESS",
        "BAD_SUPPLY",
        "ULTIMATE_WEAPONS",
        "BAD_DETECTION",
        "BAD_STEALTH",
        "BAD_FUEL",
        "CTRL_EXTINCT",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "SUCCULENT_BARNACLES_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "FORTRESS_SPECIAL",
        "GAIA_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "KRAKEN_IN_THE_ICE_SPECIAL",
        "PLC_ALLIED_REPAIR",
        "PLC_CONFEDERATION",
        "PLC_ENGINEERING",
        "PLC_EXPLORATION",
        "PLC_LIBERTY",
        "PLC_MARINE_RECRUITMENT",
    ],
    dislikes=[
        "BLD_INDUSTRY_CENTER",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_STOCKPILING_CENTER",
        "BLD_MEGALITH",
        "BLD_SPACE_ELEVATOR",
        "BLD_GENOME_BANK",
        "BLD_COLLECTIVE_NET",
        "BLD_GATEWAY_VOID",
        "BLD_ENCLAVE_VOID",
        "BLD_HYPER_DAM",
        "PLC_BUREAUCRACY",
        "PLC_CENTRALIZATION",
        "PLC_CONFORMANCE",
        "PLC_DIVINE_AUTHORITY",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *BAD_RESEARCH,
        *BAD_INFLUENCE,
        *BAD_STOCKPILE,
        *BAD_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_MISIORLA"),
        *BAD_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *ULTIMATE_WEAPONS,
        *BAD_DETECTION,
        *BAD_STEALTH,
        *BAD_FUEL,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TOXIC_STANDARD_EP,
    graphic="icons/species/misiorla.png",
)



import freeOrionAIInterface as fo
from logging import debug, error, info

import FleetUtilsAI
import MoveUtilsAI
import PlanetUtilsAI
from AIDependencies import INVALID_ID, MINIMUM_GUARD_DISTANCE_TO_HOME_SYSTEM
from aistate_interface import get_aistate
from common.fo_typing import SystemId
from EnumsAI import MissionType
from freeorion_tools import get_fleet_position, get_partial_visibility_turn
from target import TargetSystem
from universe.system_network import get_neighbors, systems_connected

graph_flags = set()
border_unexplored_system_ids = set()


def get_current_exploration_info():
    """Returns ([current target list], [available scout list])."""
    fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.EXPLORATION)
    available_scouts = []
    already_covered = set()
    aistate = get_aistate()
    for fleet_id in fleet_ids:
        fleet_mission = aistate.get_fleet_mission(fleet_id)
        if not fleet_mission.type:
            available_scouts.append(fleet_id)
        else:
            if fleet_mission.type == MissionType.EXPLORATION:
                already_covered.add(fleet_mission.target.id)
                if not fleet_mission.target:
                    debug("problem determining existing exploration target systems")
                else:
                    debug("found existing exploration target: %s" % fleet_mission.target)
    debug("Current exploration targets: %s" % already_covered)
    debug("Available scout fleets: %s" % available_scouts)
    return list(already_covered), available_scouts


def assign_scouts_to_explore_systems():  # noqa: C901
    # TODO: use Graph Theory to explore closest systems
    universe = fo.getUniverse()
    capital_sys_id = PlanetUtilsAI.get_capital_sys_id()
    # order fleets to explore
    if not border_unexplored_system_ids or (capital_sys_id == INVALID_ID):
        return
    exp_systems_by_dist = sorted((universe.linearDistance(capital_sys_id, x), x) for x in border_unexplored_system_ids)
    debug(
        "Exploration system considering following system-distance pairs:\n  %s"
        % ("\n  ".join("%3d: %5.1f" % (sys_id, dist) for (dist, sys_id) in exp_systems_by_dist))
    )
    explore_list = [sys_id for dist, sys_id in exp_systems_by_dist]

    already_covered, available_scouts = get_current_exploration_info()

    debug("Explorable system IDs: %s" % explore_list)
    debug("Already targeted: %s" % already_covered)
    aistate = get_aistate()
    needs_vis = aistate.misc.setdefault("needs_vis", [])
    check_list = aistate.needsEmergencyExploration + needs_vis + explore_list
    if INVALID_ID in check_list:  # shouldn't normally happen, unless due to bug elsewhere
        for sys_list, name in [
            (aistate.needsEmergencyExploration, "aistate.needsEmergencyExploration"),
            (needs_vis, "needs_vis"),
            (explore_list, "explore_list"),
        ]:
            if INVALID_ID in sys_list:
                error("INVALID_ID found in " + name, exc_info=True)
    # emergency coverage can be due to invasion detection trouble, etc.
    debug("Check list: %s" % check_list)
    needs_coverage = [sys_id for sys_id in check_list if sys_id not in already_covered and sys_id != INVALID_ID]
    debug("Needs coverage: %s" % needs_coverage)

    debug(
        "Available scouts & AIstate locs: %s"
        % [(x, aistate.fleetStatus.get(x, {}).get("sysID", INVALID_ID)) for x in available_scouts]
    )
    debug("Available scouts & universe locs: %s" % [(x, universe.getFleet(x).systemID) for x in available_scouts])
    if not needs_coverage or not available_scouts:
        return

    # clean up targets which can not or don't need to be scouted
    for sys_id in list(needs_coverage):
        if sys_id not in explore_list:  # doesn't necessarily need direct visit
            if universe.getVisibility(sys_id, fo.empireID()) >= fo.visibility.partial:
                # already got visibility; remove from visit lists and skip
                if sys_id in needs_vis:
                    del needs_vis[needs_vis.index(sys_id)]
                if sys_id in aistate.needsEmergencyExploration:
                    del aistate.needsEmergencyExploration[aistate.needsEmergencyExploration.index(sys_id)]
                debug("system id %d already currently visible; skipping exploration" % sys_id)
                needs_coverage.remove(sys_id)
                continue

        # skip systems threatened by monsters
        sys_status = aistate.systemStatus.setdefault(sys_id, {})
        if not aistate.character.may_explore_system(sys_status.setdefault("monsterThreat", 0)) or (
            fo.currentTurn() < 20 and aistate.systemStatus[sys_id]["monsterThreat"] > 0
        ):
            debug(
                "Skipping exploration of system %d due to Big Monster, threat %d"
                % (sys_id, aistate.systemStatus[sys_id]["monsterThreat"])
            )
            needs_coverage.remove(sys_id)
            continue

    # find the jump distance for all possible scout-system pairings
    options = []
    available_scouts = set(available_scouts)
    for fleet_id in available_scouts:
        start = TargetSystem(get_fleet_position(fleet_id))
        for sys_id in needs_coverage:
            target = TargetSystem(sys_id)
            path = MoveUtilsAI.can_travel_to_system(fleet_id, start, target, ensure_return=True)
            if not path:
                continue
            num_jumps = len(path) - 1  # -1 as path contains the original system
            options.append((num_jumps, fleet_id, sys_id))

    # Apply a simple, greedy heuristic to match scouts to nearby systems:
    # Always choose the shortest possible path from the remaining scout-system pairing.
    # This is clearly not optimal in the general case but it works well enough for now.
    # TODO: Consider using a more sophisticated assignment algorithm
    options.sort()
    while options:
        debug("Remaining options: %s" % options)
        _, fleet_id, sys_id = options[0]
        fleet_mission = aistate.get_fleet_mission(fleet_id)
        target = TargetSystem(sys_id)
        info("Sending fleet %d to explore %s" % (fleet_id, target))
        fleet_mission.set_target(MissionType.EXPLORATION, target)
        options = [option for option in options if option[1] != fleet_id and option[2] != sys_id]
        available_scouts.remove(fleet_id)
        needs_coverage.remove(sys_id)

    debug("Exploration assignment finished.")
    debug("Unassigned scouts: %s" % available_scouts)
    debug("Unassigned exploration targets: %s" % needs_coverage)


def follow_vis_system_connections(start_system_id, home_system_id):  # noqa: C901
    universe = fo.getUniverse()
    empire_id = fo.empireID()
    exploration_list = [start_system_id]
    aistate = get_aistate()
    while exploration_list:
        cur_system_id = exploration_list.pop()
        if cur_system_id in graph_flags:
            continue
        graph_flags.add(cur_system_id)
        system = universe.getSystem(cur_system_id)
        if cur_system_id in aistate.visBorderSystemIDs:
            pre_vis = "a border system"
        elif cur_system_id in aistate.visInteriorSystemIDs:
            pre_vis = "an interior system"
        else:
            pre_vis = "an unknown system"
        system_header = "*** system %s;" % system
        if fo.currentTurn() < 50:
            visibility_turn_list = sorted(
                universe.getVisibilityTurnsMap(cur_system_id, empire_id).items(), key=lambda x: x[0].numerator
            )
            visibility_info = ", ".join(f"{vis.name}: {turn}" for vis, turn in visibility_turn_list)
            debug(f"{system_header} previously {pre_vis}. Visibility per turn: {visibility_info} ")
            status_info = []
        else:
            status_info = [system_header]

        has_been_visible = get_partial_visibility_turn(cur_system_id) > 0
        is_connected = systems_connected(cur_system_id, home_system_id)
        status_info.append("    -- is%s partially visible" % ("" if has_been_visible else " not"))
        status_info.append("    -- is%s visibly connected to homesystem" % ("" if is_connected else " not"))
        if has_been_visible:
            sys_status = aistate.systemStatus.setdefault(cur_system_id, {})
            aistate.visInteriorSystemIDs.add(cur_system_id)
            aistate.visBorderSystemIDs.discard(cur_system_id)
            neighbors = get_neighbors(cur_system_id)
            sys_status.setdefault("neighbors", set()).update(neighbors)
            if neighbors:
                status_info.append(" -- has neighbors %s" % sorted(neighbors))
                for sys_id in neighbors:
                    if sys_id not in aistate.exploredSystemIDs:
                        aistate.unexploredSystemIDs.add(sys_id)
                    if (sys_id not in graph_flags) and (sys_id not in aistate.visInteriorSystemIDs):
                        aistate.visBorderSystemIDs.add(sys_id)
                        exploration_list.append(sys_id)
        if fo.currentTurn() < 50:
            debug("\n".join(status_info))
            debug("----------------------------------------------------------")


def update_explored_systems():
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    obs_lanes = empire.obstructedStarlanes()
    obs_lanes_list = [el for el in obs_lanes]  # should result in list of tuples (sys_id1, sys_id2)
    if obs_lanes_list:
        debug("Obstructed starlanes are: %s" % ", ".join("{}-{}".format(*item) for item in obs_lanes_list))
    else:
        debug("No obstructed Starlanes")
    newly_explored = []
    still_unexplored = []
    aistate = get_aistate()
    for sys_id in list(aistate.unexploredSystemIDs):
        # consider making determination according to visibility rather than actual visit,
        # which I think is what empire.hasExploredSystem covers (Dilvish-fo)
        if empire.hasExploredSystem(sys_id):
            aistate.unexploredSystemIDs.discard(sys_id)
            aistate.exploredSystemIDs.add(sys_id)
            system = universe.getSystem(sys_id)
            debug("Moved system %s from unexplored list to explored list" % system)
            border_unexplored_system_ids.discard(sys_id)
            newly_explored.append(sys_id)
        else:
            still_unexplored.append(sys_id)

    neighbor_list = []
    dummy = []
    for id_list, next_list in [(newly_explored, neighbor_list), (neighbor_list, dummy)]:
        for sys_id in id_list:
            neighbors = get_neighbors(sys_id)
            for neighbor_id in neighbors:
                # when it matters, unexplored will be smaller than explored
                if neighbor_id not in aistate.unexploredSystemIDs:
                    next_list.append(neighbor_id)

    for sys_id in still_unexplored:
        neighbors = get_neighbors(sys_id)
        if any(nid in aistate.exploredSystemIDs for nid in neighbors):
            border_unexplored_system_ids.add(sys_id)
    return newly_explored


def request_emergency_exploration(system_id: SystemId):
    aistate = get_aistate()
    if system_id not in aistate.needsEmergencyExploration:
        aistate.needsEmergencyExploration.append(system_id)


def system_could_have_unknown_stationary_guard(system_id: SystemId) -> bool:
    """Return True if the system may have spawned stationary guards.

    A stationary guard is defined as immobile monster fleets spawned at game start.
    If this function indicates that there is no such guard, there still may be other threats.
    """
    # We do not play around invisible guards, so if system was visible at some point,
    # there should not be a stationary guard there
    system_was_visible = get_partial_visibility_turn(system_id) > 0
    if system_was_visible:
        return False

    # Universe setup settings may forbid guards
    if fo.getGalaxySetupData().monsterFrequency == fo.galaxySetupOptionMonsterFreq.none:
        return False

    # Stationary guards require some distance to the home system to be spawned
    home_system = PlanetUtilsAI.get_capital_sys_id()
    jump_distance_to_home_system = fo.getUniverse().jumpDistance(system_id, home_system)
    if jump_distance_to_home_system < MINIMUM_GUARD_DISTANCE_TO_HOME_SYSTEM:
        return False

    # No indicator that there isn't a stationary guard
    return True



"""


===== Columns =====



Process finished with exit code 0
"""
from common.print_utils import Number, Sequence, Table, Text, as_columns, dict_to_table

EXPECTED_ITEMS_MULTIPLE_COLUMNS = """a   c
b   d"""

EXPECTED_ITEMS_NOT_MULTIPLE_COLUMNS = """a   c
b    """

EXPECTED_SIMPLE_TABLE = """Wooho
============================================================================
| name*            | value*     | zzz                 | zzzzzzzzzzzzzzzzzz |
============================================================================
| hello            |  144444.00 | a, b, c, f, f, f, f | a                  |
| Plato aaβ III    |      21.00 | d, e                | a                  |
| Plato β III      |      21.00 | d, e                | a                  |
| Plato B III      |      21.00 | d                   | a                  |
| Plato Bddddd III |      21.00 | d                   | a                  |
| Plato III        |      21.00 | d                   | a                  |
----------------------------------------------------------------------------
*name   Name for first column
*value  VValue"""

EXPECTED_EMPTY_TABLE = """Wooho
=============================================
| name* | value* | zzz | zzzzzzzzzzzzzzzzzz |
=============================================
---------------------------------------------
*name   Name for first column
*value  VValue"""


def test_as_columns_with_items_multiple_columns():
    out = as_columns(["a", "b", "c", "d"], 2)
    assert out == EXPECTED_ITEMS_MULTIPLE_COLUMNS


def test_as_columns_with_items_not_multiple_columns():
    out = as_columns(["a", "b", "c"], 2)
    assert out == EXPECTED_ITEMS_NOT_MULTIPLE_COLUMNS


def make_table():
    t = Table(
        Text("name", description="Name for first column"),
        Number("value", description="VValue"),
        Sequence("zzz"),
        Sequence("zzzzzzzzzzzzzzzzzz"),
        table_name="Wooho",
    )
    t.add_row("hello", 144444, "abcffff", "a")
    t.add_row("Plato aa\u03b2 III", 21, "de", "a")
    t.add_row("Plato \u03b2 III", 21, "de", "a")
    t.add_row("Plato B III", 21, "d", "a")
    t.add_row("Plato Bddddd III", 21, "d", "a")
    t.add_row("Plato III", 21, "d", "a")
    return t


def test_table_is_converted_to_str():
    table = make_table()
    assert str(table) == EXPECTED_SIMPLE_TABLE


def test_empty_table():
    empty = Table(
        Text("name", description="Name for first column"),
        Number("value", description="VValue"),
        Sequence("zzz"),
        Sequence("zzzzzzzzzzzzzzzzzz"),
        table_name="Wooho",
    )
    assert str(empty) == EXPECTED_EMPTY_TABLE


def test_number_column():
    field = Number("name", placeholder="-")
    assert field.make_cell_string(0) == "-"


def test_total_is_calculated():
    table = Table(
        Number("A", precession=0, total=True),
        Number("B", precession=0, total=True),
    )
    table.add_row(1, 10)
    table.add_row(2, 20)
    assert list(table) == [
        "============",
        "| A  | B   |",
        "============",
        "|  1 |  10 |",
        "|  2 |  20 |",
        "============",
        "|  3 |  30 |",
        "============",
    ]


def test_hide_header():
    table = Table(
        Number("A", precession=0),
        Number("B", precession=0),
        hide_header=True,
    )
    table.add_row(1, 10)
    table.add_row(2, 20)
    assert list(table) == [
        "============",
        "|  1 |  10 |",
        "|  2 |  20 |",
        "------------",
    ]


def test_hide_header_with_no_rows():
    table = Table(
        Number("A", precession=0),
        Number("B", precession=0),
        hide_header=True,
    )
    assert list(table) == [
        "=========",
        "---------",
    ]


def test_dict_to_table():
    data = {"1": 10, "2": 20}
    table = dict_to_table(data)

    assert list(table) == [
        "==========",
        "| 1 | 10 |",
        "| 2 | 20 |",
        "----------",
    ]


def test_adding_notes():
    table = Table(
        Number("A", precession=0),
        Number("B", precession=0),
        hide_header=True,
    )
    table.add_row(1, 10)
    table.add_row(2, 20, note="Hello")
    table.add_row(3, 30, note="  World  ")
    assert list(table) == [
        "============",
        "|  1 |  10 |",
        "|  2 |  20 |  Hello",
        "|  3 |  30 |  World",
        "------------",
    ]



import freeOrionAIInterface as fo
from unittest.mock import patch

import pytest

from AIDependencies import INVALID_ID
from universe import system_network


class TestUniverse:
    connected_systems = {
        1: {2, 7},
        2: {1, 3},
        3: {2, 4, 5, 6},
        4: {3, 5},
        5: {3, 4, 6},
        6: {3, 5, 7},
        7: {1, 6, 8},
        8: {7, 9},
        9: {8, 10},
        10: {9},
        11: {12},
        12: {11},
    }

    # fo interface asks for an empire_id, so we need two arguments as well
    def getImmediateNeighbors(self, sys_id, _):
        return TestUniverse.connected_systems.get(sys_id, {})

    @staticmethod
    def systemsConnected(sx, sy, _):
        return sx in system_network.systems_connected_to_system(sy)


@pytest.fixture(autouse=True)
def connected_systems(monkeypatch):
    monkeypatch.setattr(fo, "getUniverse", TestUniverse)


def test_connections():
    """
    Test that the graph is bi-directional.
    This simplifies the implementation of TestUniverse and makes it easier to determine the expected results.
    """
    for i1, lanes in TestUniverse.connected_systems.items():
        for i2 in lanes:
            assert i1 in TestUniverse.connected_systems[i2]


@pytest.mark.parametrize(
    ("sys_id", "expected"),
    (
        (1, {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}),
        (4, {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}),
        (12, {11, 12}),
        (INVALID_ID, set()),
    ),
)
def test_systems_connected_to_system(sys_id, expected):
    assert system_network.systems_connected_to_system(sys_id) == expected


@pytest.mark.parametrize(
    ("sys_id", "n", "expected"),
    (
        (1, 0, {1}),
        (1, 1, {1, 2, 7}),
        (1, 2, {1, 2, 3, 6, 7, 8}),
        (1, 3, {1, 2, 3, 4, 5, 6, 7, 8, 9}),
        (1, 4, {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}),
        (1, 5, {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}),
        (10, 1, {9, 10}),
        (10, 2, {8, 9, 10}),
        (10, 3, {7, 8, 9, 10}),
        (10, 4, {1, 6, 7, 8, 9, 10}),
        (10, 5, {1, 2, 3, 5, 6, 7, 8, 9, 10}),
        (11, 5, {11, 12}),
    ),
)
def test_within_n_jumps(sys_id, n, expected):
    assert system_network.within_n_jumps(sys_id, n) == expected


@pytest.mark.parametrize(
    ("s1", "s2", "expected"), ((1, 2, True), (1, 12, False), (11, 1, False), (11, 12, True), (11, 13, False))
)
def test_systems_connected(s1, s2, expected, monkeypatch):
    with patch("universe.system_network.get_capital_sys_id") as get_capital_sys_id:
        get_capital_sys_id.return_value = 1
        assert system_network.systems_connected(s1, s2) == expected
        assert system_network.systems_connected(s2, s1) == expected



import freeOrionAIInterface as fo
from logging import error, warning
from typing import Optional, Union

import AIDependencies
from AIDependencies import ALL_EMPIRES
from EnumsAI import EmpireMeters
from freeorion_tools import get_species_stealth
from freeorion_tools.caching import cache_for_current_turn


def default_empire_detection_strength():
    # TODO doublecheck typical AI research times for Radar, for below default value
    return (
        AIDependencies.DETECTION_TECH_STRENGTHS["SPY_DETECT_1"]
        if fo.currentTurn() < 40
        else AIDependencies.DETECTION_TECH_STRENGTHS["SPY_DETECT_2"]
    )


@cache_for_current_turn
def get_empire_detection(empire_id: int) -> float:
    """
    Returns the detection strength for the provided empire ID.

    If passed the ALL_EMPIRES ID, then it returns the max detection strength across
    all empires except for the current AI's empire.
    """
    if empire_id == ALL_EMPIRES:
        return get_max_empire_detection(fo.allEmpireIDs())

    empire = fo.getEmpire(empire_id)
    if empire:
        return empire.getMeter(EmpireMeters.DETECTION_STRENGTH).initial
    else:
        warning("AI failed to retrieve empire ID %d, in game with %d empires." % (empire_id, len(fo.allEmpireIDs())))
        return default_empire_detection_strength()


def get_max_empire_detection(empire_list: Union[list[int], "fo.IntVec"]) -> float:
    """
    Returns the max detection strength across all empires except for the current AI's empire.

    :param empire_list: list of empire IDs
    :return: max detection strength of provided empires, excluding the current self empire
    """
    max_detection = 0
    for this_empire_id in empire_list:
        if this_empire_id != fo.empireID():
            max_detection = max(max_detection, get_empire_detection(this_empire_id))
    return max_detection


def colony_detectable_by_empire(
    planet_id: int,
    species_name: Optional[str] = None,
    empire: Union[int, list[int]] = ALL_EMPIRES,
    future_stealth_bonus: int = 0,
    default_result: bool = True,
) -> bool:
    """
    Predicts if a planet/colony is/will-be detectable by an empire.

    The passed empire value can be a single empire ID or a list of empire IDs.  If passed ALL_empires, will use the list
    of all empires.  When using a list of empires (or when passed ALL_EMPIRES), the present empire is excluded.  To
    check for the present empire the current empire ID must be passed as a simple int.  Ignores current ownership of the
    planet unless the passed empire value is a simple int matching fo.empireID(), because in most cases we are concerned
    about (i) visibility to us of a planet we do not own, or (ii) visibility by enemies of a planet we own or expect to
    own at the time we are making the visibility projection for (even if they may own it right now).  The only case
    where current ownership matters is when we are considering whether to abort a colonization mission, which might be
    for a planet we already own.

    :param planet_id: required, the planet of concern
    :param species_name: will override the existing planet species if provided
    :param empire: empire ID (or list of empire IDs) whose detection ability is of concern
    :param future_stealth_bonus: can specify a projected future stealth bonus, such as from a stealth tech
    :param default_result: generally for offensive assessments should be False, for defensive should be True
    :return: whether the planet is predicted to be detectable

    """
    # The future_stealth_bonus can be used if the AI knows it has researched techs that would grant a stealth bonus to
    # the planet once it was colonized/captured

    planet = fo.getUniverse().getPlanet(planet_id)
    if not planet:
        error("Couldn't retrieve planet ID %d." % planet_id)
        return default_result
    if species_name is None:
        species_name = planet.speciesName

    # in case we are checking about aborting a colonization mission
    if empire == fo.empireID() and planet.ownedBy(empire):
        return True

    # could just check stealth meter, but this approach might allow us to plan ahead a bit even if the planet
    # is temporarily stealth boosted by temporary effects like ion storm
    planet_stealth = AIDependencies.BASE_PLANET_STEALTH
    if planet.specials:
        planet_stealth += max([AIDependencies.STEALTH_SPECIAL_STRENGTHS.get(_spec, 0) for _spec in planet.specials])
    # TODO: check planet buildings for stealth bonuses

    # if the planet already has an existing stealth special, then the most common situation is that it would be
    # overlapping with or superseded by the future_stealth_bonus, not additive with it.
    planet_stealth = max(planet_stealth, AIDependencies.BASE_PLANET_STEALTH + future_stealth_bonus)
    species_stealth_mod = get_species_stealth(species_name)
    total_stealth = planet_stealth + species_stealth_mod
    if isinstance(empire, int):
        empire_detection = get_empire_detection(empire)
    else:
        empire_detection = get_max_empire_detection(empire)
    return total_stealth < empire_detection



import freeorion as fo
import random

import names
import options

# for starname modifiers
stargroup_words = []
stargroup_chars = []
stargroup_modifiers = []
greek_letters = [
    "Alpha",
    "Beta",
    "Gamma",
    "Delta",
    "Epsilon",
    "Zeta",
    "Eta",
    "Theta",
    "Iota",
    "Kappa",
    "Lambda",
    "Mu",
    "Nu",
    "Xi",
    "Omicron",
    "Pi",
    "Rho",
    "Sigma",
    "Tau",
    "Upsilon",
    "Phi",
    "Chi",
    "Psi",
    "Omega",
]


def random_star_name():
    """
    Returns a random star system name.
    """
    return "".join(random.sample(names.consonants + names.vowels, 3)).upper() + "-" + str(random.randint(1000, 9999))


# used in clustering
def recalc_centers(ctrs, points, assignments):
    tallies = [[[], []] for _ in ctrs]
    for index_p, index_ctr in enumerate(assignments):
        tallies[index_ctr][0].append(points[index_p][0])
        tallies[index_ctr][1].append(points[index_p][1])
    for index_ctr, tally in enumerate(tallies):
        num_here = len(tally[0])
        if num_here == 0:
            pass  # leave ctr unchanged if no points assigned to it last round
        else:
            ctrs[index_ctr] = [float(sum(tally[0])) / num_here, float(sum(tally[1])) / num_here]


# used in clustering
def assign_clusters(points, ctrs):
    assignments = []
    for point in points:
        best_dist_sqr = 1e20
        best_ctr = 0
        for index_ctr, ctr in enumerate(ctrs):
            this_dist_sqr = (ctr[0] - point[0]) ** 2 + (ctr[1] - point[1]) ** 2  # TODO use math.hypot, max
            if this_dist_sqr < best_dist_sqr:
                best_dist_sqr = this_dist_sqr
                best_ctr = index_ctr
        assignments.append(best_ctr)
    return assignments


def cluster_stars(positions, num_star_groups):
    """
    Returns a list, same size as positions argument, containing indices from 0 to num_star_groups.
    """

    if num_star_groups > len(positions):
        return [[pos] for pos in positions]
    centers = [[pos[0], pos[1]] for pos in random.sample(positions, num_star_groups)]
    all_coords = [(pos[0], pos[1]) for pos in positions]
    clusters = [[], []]
    clusters[0] = assign_clusters(all_coords, centers)  # assign clusters based on init centers
    old_c = 0
    num_convergence_loops = 1  # if full convergence is deemed important then use a higher humber here
    for loop in range(num_convergence_loops):  # main loop to try getting some convergence of center assignments
        recalc_centers(centers, all_coords, clusters[old_c])  # get new centers
        clusters[1 - old_c] = assign_clusters(all_coords, centers)  # assign clusters based on new centers
        if clusters[0] == clusters[1]:
            break  # stop iterating if no change in cluster assignments
        old_c = 1 - old_c
    else:
        if loop > 0:  # if here at loop 0, then didn't try for convergence
            print("falling through system clustering iteration loop without convergence")
    return clusters[1 - old_c]


def check_deep_space(group_list, star_type_assignments, planet_assignments):
    deep_space = []
    not_deep = []
    for systemxy in group_list:
        if star_type_assignments.get(systemxy, fo.starType.noStar) != fo.starType.noStar:
            not_deep.append(systemxy)
        else:
            if len(planet_assignments.get(systemxy, [])) > 0:
                not_deep.append(systemxy)
            else:
                deep_space.append(systemxy)
    return not_deep, deep_space


def name_group(group_list, group_name, star_type_assignments, planet_assignments):
    group_size = len(group_list)
    if group_size == 1:
        return [(group_list[0], group_name)]
    modifiers = list(stargroup_modifiers)  # copy the list so we can safely add to it if needed
    not_deep, deep_space = check_deep_space(group_list, star_type_assignments, planet_assignments)
    these_systems = not_deep + deep_space  # so that unnamed deep space will get the later star group modifiers
    while len(modifiers) < group_size:  # emergency fallback
        trial_mod = (
            random.choice(stargroup_modifiers)
            + " "
            + "".join(random.sample(names.consonants + names.vowels, 3)).upper()
        )
        if trial_mod not in modifiers:
            modifiers.append(trial_mod)
    if options.POSTFIX_STARGROUP_MODIFIERS:
        name_list = [group_name + " " + modifier for modifier in modifiers]
    else:
        name_list = [modifier + " " + group_name for modifier in modifiers]
    return zip(these_systems, name_list)


def name_star_systems(system_list):  # noqa: C901
    # choose star types and planet sizes, before choosing names, so naming can have special handling of Deep Space
    star_type_assignments = {}
    planet_assignments = {}
    position_list = []
    for system in system_list:
        star_type = fo.sys_get_star_type(system)
        systemxy = fo.get_pos(system)
        star_type_assignments[systemxy] = star_type
        planet_assignments[systemxy] = fo.sys_get_planets(system)
        position_list.append(systemxy)

    # will name name a portion of stars on a group basis, where the stars of each group share the same base star name,
    # suffixed by different (default greek) letters or characters (options at top of file)
    star_name_map = {}
    star_names = names.get_name_list("STAR_NAMES")
    group_names = names.get_name_list("STAR_GROUP_NAMES")
    potential_group_names = []
    individual_names = []
    stargroup_words[:] = names.get_name_list("STAR_GROUP_WORDS")
    stargroup_chars[:] = names.get_name_list("STAR_GROUP_CHARS")
    stargroup_modifiers[:] = [stargroup_words, stargroup_chars][options.STAR_GROUPS_USE_CHARS]
    for starname in star_names:
        if len(starname) > 6:  # if starname is long, don't allow it for groups
            individual_names.append(starname)
            continue
        # any names that already have a greek letter in them can only be used for individual stars, not groups
        for namepart in starname.split():
            if namepart in greek_letters:
                individual_names.append(starname)
                break
        else:
            potential_group_names.append(starname)

    if not potential_group_names:
        potential_group_names.append("XYZZY")

    # ensure at least a portion of galaxy gets individual starnames
    num_systems = len(system_list)
    choice = num_systems >= options.NAMING_LARGE_GALAXY_SIZE
    target_indiv_ratio = [options.TARGET_INDIV_RATIO_SMALL, options.TARGET_INDIV_RATIO_LARGE][choice]
    # TODO improve the following calc to be more likely to hit target_indiv_ratio if more or less than
    # 50% potential_group_names used for groups
    num_individual_stars = int(
        max(
            min(num_systems * target_indiv_ratio, len(individual_names) + int(0.5 * len(potential_group_names))),
            num_systems - 0.8 * len(stargroup_modifiers) * (len(group_names) + int(0.5 * len(potential_group_names))),
        )
    )
    star_group_size = 1 + int(
        (num_systems - num_individual_stars) / (max(1, len(group_names) + int(0.5 * len(potential_group_names))))
    )
    # make group size a bit bigger than min necessary, at least a trio
    star_group_size = max(3, star_group_size)
    num_star_groups = 1 + int(num_systems / star_group_size)  # initial value

    # first cluster all systems, then remove some to be individually named (otherwise groups can have too many
    # individually named systems in their middle).  First remove any that are too small (only 1 or 2 systems).
    # The clusters with the most systems are generally the most closely spaced, and though they might make good
    # logical candidates for groups, their names are then prone to overlapping on the galaxy map, so after removing
    # small groups, remove the groups with the most systems.
    random.shuffle(position_list)  # just to be sure it is randomized
    init_cluster_assgts = cluster_stars(position_list, num_star_groups)
    star_groups = {}
    for index_pos, index_group in enumerate(init_cluster_assgts):
        systemxy = position_list[index_pos]
        star_groups.setdefault(index_group, []).append(systemxy)
    indiv_systems = []

    # remove groups with only one non-deep-system
    for groupindex, group_list in list(star_groups.items()):
        max_can_transfer = len(potential_group_names) - len(star_groups) + len(individual_names) - len(indiv_systems)
        if max_can_transfer <= 0:
            break
        elif max_can_transfer <= len(group_list):
            continue
        not_deep, deep_space = check_deep_space(group_list, star_type_assignments, planet_assignments)
        if len(not_deep) > 1:
            continue
        for systemxy in star_groups[groupindex]:
            indiv_systems.append(systemxy)
        del star_groups[groupindex]

    # remove tiny groups
    group_sizes = [(len(group), index) for index, group in star_groups.items()]
    group_sizes.sort()
    while len(indiv_systems) < num_individual_stars and len(group_sizes) > 0:
        groupsize, groupindex = group_sizes.pop()
        max_can_transfer = len(potential_group_names) - len(star_groups) + len(individual_names) - len(indiv_systems)
        if (max_can_transfer <= 0) or (groupsize > 2):
            break
        if max_can_transfer <= groupsize:
            continue
        for systemxy in star_groups[groupindex]:
            indiv_systems.append(systemxy)
        del star_groups[groupindex]

    # remove largest (likely most compact) groups
    while len(indiv_systems) < num_individual_stars and len(group_sizes) > 0:
        groupsize, groupindex = group_sizes.pop(-1)
        max_can_transfer = len(potential_group_names) - len(star_groups) + len(individual_names) - len(indiv_systems)
        if max_can_transfer <= 0:
            break
        if max_can_transfer <= groupsize:
            continue
        for systemxy in star_groups[groupindex]:
            indiv_systems.append(systemxy)
        del star_groups[groupindex]

    num_star_groups = len(star_groups)
    num_individual_stars = len(indiv_systems)
    random.shuffle(potential_group_names)
    random.shuffle(individual_names)
    random.shuffle(group_names)
    num_for_indiv = min(
        max(len(potential_group_names) // 2, num_individual_stars + 1 - len(individual_names)),
        len(potential_group_names),
    )
    individual_names.extend(potential_group_names[:num_for_indiv])
    group_names.extend(potential_group_names[num_for_indiv:])

    # print "sampling for %d indiv names from list of %d total indiv names" % (
    #     num_individual_stars, len(individual_names))
    indiv_name_sample = random.sample(individual_names, num_individual_stars)
    # indiv_name_assignments = zip([(pos.x, pos.y) for pos in position_list[:num_individual_stars]], indiv_name_sample)
    star_name_map.update(zip(indiv_systems, indiv_name_sample))
    # print "sampling for %d group names from list of %d total group names"%(num_star_groups, len(group_names))
    if len(group_names) < num_star_groups:
        group_names.extend([names.random_name(6) for _ in range(num_star_groups - len(group_names))])
    group_name_sample = random.sample(group_names, num_star_groups)
    for index_group, group_list in enumerate(sorted(star_groups.values())):
        star_name_map.update(
            name_group(group_list, group_name_sample[index_group], star_type_assignments, planet_assignments)
        )

    # assign names from star_name_map to star systems
    for system in system_list:
        fo.set_name(system, star_name_map.get(fo.get_pos(system), "") or random_star_name())



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ORGANIC_STRC",
    description="CON_ORGANIC_STRC_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=280 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["GRO_SYMBIOTIC_BIO", "CON_FRC_ENRG_STRC"],
    unlock=Item(type=UnlockBuilding, name="BLD_TRANSFORMER"),
    graphic="icons/tech/transforming_structures.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SPINAL_ANTIMATTER",
    description="SHP_SPINAL_ANTIMATTER_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=300 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["PRO_ZERO_GEN"],
    unlock=Item(type=UnlockShipPart, name="SR_SPINAL_ANTIMATTER"),
    graphic="icons/ship_parts/spinal_antimatter.png",
)



from focs._effects import (
    EffectsGroup,
    InSystem,
    LocalCandidate,
    OwnedBy,
    SetStructure,
    Ship,
    Source,
    Stationary,
    Target,
    Turn,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ADV_DAM_CONT",
    description="SHP_ADV_DAM_CONT_DESC",
    short_description="STRUCTURE_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=160 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_DAMAGE_CONTROL_PART_TECHS"],
    prerequisites=["SHP_FLEET_REPAIR"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship
            & OwnedBy(empire=Source.Owner)
            & InSystem()
            & Stationary
            & Turn(low=LocalCandidate.System.LastTurnBattleHere + 1),
            effects=SetStructure(value=Value + (Target.MaxStructure / 10)),
        )
    ],
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_NEUTRONIUM_EXTRACTION",
    description="PRO_NEUTRONIUM_EXTRACTION_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=240 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_ZERO_GEN", "LRN_STELLAR_TOMOGRAPHY"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_NEUTRONIUM_EXTRACTOR"),
        Item(type=UnlockBuilding, name="BLD_NEUTRONIUM_FORGE"),
        Item(type=UnlockShipPart, name="AR_NEUTRONIUM_PLATE"),
    ],
    graphic="icons/tech/neutronium_extraction.png",
)



import freeOrionAIInterface as fo

from AIDependencies import INVALID_ID
from common.fo_typing import FleetId, SystemId


def get_fleet_position(fleet_id: FleetId) -> SystemId:
    """Return system where fleet is located or moving to."""
    fleet = fo.getUniverse().getFleet(fleet_id)
    return fleet.systemID if fleet.systemID != INVALID_ID else fleet.nextSystemID



import os
import sys

import streamlit as st

from tabs.policies import plot_policy_adoptions
from tabs.stats import draw_plots

current_dir = os.path.dirname(__file__)
common = os.path.join(current_dir, "..")
assert os.path.exists(common)
sys.path.append(common)

from charting.log_parser.collect_data import get_ais_data  # noqa: E402, need to inject common dir before import


def setup_layout():
    st.set_page_config(layout="wide")


setup_layout()

plots_tab, policies_tab = st.tabs(["Stats", "Policies"])

with plots_tab:
    draw_plots(get_ais_data())

with policies_tab:
    plot_policy_adoptions(get_ais_data())



"""
This module contains FreeOrion types.

Since a lot of object are operated by their ID
it is hard to distinguish one int from the another.
"""

from typing import NewType

_UniverseObjectId = NewType("UniverseObjectId", int)
ObjectId = NewType("ObjectId", _UniverseObjectId)
PlanetId = NewType("PlanetId", _UniverseObjectId)
FleetId = NewType("FleetId", _UniverseObjectId)
SystemId = NewType("SystemId", _UniverseObjectId)
ShipId = NewType("ShipId", _UniverseObjectId)
BuildingId = NewType("BuildingId", _UniverseObjectId)


EmpireId = NewType("EmpireId", int)
PlayerId = NewType("PlayerId", int)

BuildingName = NewType("BuildingName", str)
SpeciesName = NewType("SpeciesName", str)
SpecialName = NewType("SpecialName", str)
PartName = NewType("PartName", str)

Turn = NewType("Turn", int)

AttackDamage = NewType("AttackDamage", float)
AttackCount = NewType("AttackCount", int)



from focs._effects import (
    AddSpecial,
    DirectDistanceBetween,
    EffectsGroup,
    IsField,
    IsSource,
    LocalCandidate,
    MinimumNumberOf,
    MoveTowards,
    NoStar,
    Source,
    Star,
    Turn,
)
from focs._species import *
from species.species_macros.detection import GREAT_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import INFERNO_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import EXTREMELY_BAD_POPULATION, LIGHT_SENSITIVE
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.telepathic import HAEMAESTHETIC_DETECTION
from species.species_macros.troops import BAD_OFFENSE_TROOPS, GREAT_DEFENSE_TROOPS

Species(
    name="SP_NIGHTSIDERS",
    description="SP_NIGHTSIDERS_DESC",
    gameplay_description="SP_NIGHTSIDERS_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    tags=[
        "ORGANIC",
        "NO_INFLUENCE",
        "EXTREMELY_BAD_POPULATION",
        "GREAT_DETECTION",
        "GOOD_STEALTH",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "HAEMAESTHETIC_DETECTION",
        "BAD_OFFENSE_TROOPS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_PROTECTION",
    likes=[
        "MIMETIC_ALLOY_SPECIAL",  # because it works in the dark
        "TIDAL_LOCK_SPECIAL",  # no place like home
        "FOCUS_PROTECTION",  # good at it and like it. why not. go away. nothing to see here anyway. we'll come to you when we need your blood.
        "BLD_INTERSPECIES_ACADEMY",  # sure, show us all the species you got, maybe tasty
        "PLC_DIVINE_AUTHORITY",  # we were kind of in god's plan from the beginning ...
        "PLC_ISOLATION",
        "PLC_INSURGENCY",
        "SP_DERTHREAN",  # now following: whats on the menu – all those tasty organic non-aquatic species
        "SP_EAXAW",
        "SP_FIFTYSEVEN",
        "SP_FURTHEST",
        "SP_GYSACHE",
        "SP_HHHOH",
        "SP_HUMAN",
        "SP_LEMBALALAM",
        "SP_MISIORLA",
        "SP_MUURSH",
        "SP_PHINNERT",
        "SP_RAAAGH",
        "SP_SLEEPERS",
    ],
    dislikes=[
        "PANOPTICON_SPECIAL",  # can see well enough ourselves so no thx
        "BLD_SCANNING_FACILITY",  # same. emphasis.
        "SHIMMER_SILK_SPECIAL",  # draws unnecessary attention
        "BLD_LIGHTHOUSE",  # seriously?
        "SP_SETINON",  # cause we can feel they are there, but too small to be sucked. suckers!
        "SP_SSLITH",  # see above
        "SP_GISGUFGTHRIM",  # shell too hard and that amber sap makes their blood taste bitter
        "SP_OURBOOLS",
        "SP_SCYLIOR",  # and of course we hate all the other inedible organic species as well.
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *NO_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *EXTREMELY_BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS(
            "SP_NIGHTSIDERS"
        ),  # TODO: make like empires with lots of organic population? or just use a standard count of liked species in empire?
        *AVERAGE_SUPPLY,
        *BAD_OFFENSE_TROOPS,
        *GREAT_DEFENSE_TROOPS,
        *GREAT_DETECTION,
        *GOOD_STEALTH,
        *HAEMAESTHETIC_DETECTION(9),
        *LIGHT_SENSITIVE,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        NARROW_EP,
        # the nightsiders are so bad, they invite fire and brimstone,
        EffectsGroup(
            scope=MinimumNumberOf(
                number=1,
                sortkey=DirectDistanceBetween(Source.ID, LocalCandidate.ID),
                condition=IsField(name=["FLD_METEOR_BLIZZARD"]),
            ),
            effects=MoveTowards(speed=5, target=IsSource),
        ),
        # the nightsiders are so bad, they even repel good fields,
        EffectsGroup(
            scope=MinimumNumberOf(
                number=1,
                sortkey=DirectDistanceBetween(Source.ID, LocalCandidate.ID),
                condition=IsField(name=["FLD_NANITE_SWARM"]),
            ),
            effects=MoveTowards(speed=-5, target=IsSource),
        ),
        # they like to be kept in the dark, so either the night side,
        EffectsGroup(
            scope=IsSource,
            activation=Turn(high=0) & ~Star(type=[NoStar]),
            effects=[AddSpecial(name="TIDAL_LOCK_SPECIAL")],
        )
        # TODO: ... or a starless planet but then tidal_lock makes little sense,
    ],
    environments=INFERNO_NARROW_EP,
    graphic="icons/species/nightsiders.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import RADIATED_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import VERY_BAD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import GREAT_POPULATION
from species.species_macros.research import VERY_BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_REPLICON",
    description="SP_REPLICON_DESC",
    gameplay_description="SP_REPLICON_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "VERY_BAD_RESEARCH",
        "VERY_BAD_HAPPINESS",
        "GREAT_POPULATION",
        "ADAPTIVE",
        "PEDIA_ADAPTIVE_TITLE",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "FOCUS_INFLUENCE",
        "PANOPTICON_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "PLC_COLONIZATION",
        "PLC_COLONIALISM",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_DIVINE_AUTHORITY",
        "PLC_EXPLORATION",
        "PLC_FEUDALISM",
        "PLC_MARINE_RECRUITMENT",
        "PLC_NATIVE_APPROPRIATION",
        "PLC_POPULATION",
        "PLC_PROPAGANDA",
        "PLC_VASSALIZATION",
        "SHIMMER_SILK_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "SPICE_SPECIAL",
    ],
    dislikes=[
        # "BLD_CLONING_CENTER"       // disabled content,
        "FOCUS_RESEARCH",
        "BLD_SHIPYARD_AST",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SHIPYARD_ORG_CELL_GRO_CHAMB",
        "CRYSTALS_SPECIAL",
        "GAIA_SPECIAL",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_ENVIRONMENTALISM",
        "PLC_INDUSTRIALISM",
        "PLC_LIBERTY",
        "PLC_NO_GROWTH",
        "PLC_NO_SUPPLY",
        "PLC_TECHNOCRACY",
        "RESONANT_MOON_SPECIAL",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *VERY_BAD_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *GREAT_POPULATION,
        *VERY_BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_REPLICON"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=RADIATED_STANDARD_EP,
    graphic="icons/species/replicon.png",
)
